import fsgs
import fsui
from fsbc import settings
from launcher.i18n import gettext
from launcher.launcher_settings import LauncherSettings
from launcher.launcher_signal import LauncherSignal
from launcher.option import Option
from launcher.settings.advanced_settings_page import AdvancedSettingsPage
from launcher.settings.advanced_video_settings import AdvancedVideoSettingsPage
from launcher.settings.arcade_settings_page import ArcadeSettingsPage
from launcher.settings.audio_settings_page import AudioSettingsPage
from launcher.settings.directories_settings_page import DirectoriesSettingsPage
from launcher.settings.fs_uae_settings_page import FSUAESettingsPage
from launcher.settings.gamedatabasesettingspage import GameDatabaseSettingsPage
from launcher.settings.gameplatformssettingspage import (
    GamePlatformsSettingsPage,
)
from launcher.settings.joystick_settings_page import JoystickSettingsPage
from launcher.settings.keyboard_settings_page import KeyboardSettingsPage
from launcher.settings.language_settings_page import LanguageSettingsPage
from launcher.settings.launchersettingspage import LauncherSettingsPage
from launcher.settings.mouse_settings_page import MouseSettingsPage
from launcher.settings.netplay_settings_page import NetplaySettingsPage
from launcher.settings.plugins_settings_page import PluginsSettingsPage
from launcher.settings.scan_settings_page import ScanSettingsPage
from launcher.settings.video_settings_page import VideoSettingsPage
from launcher.settings.whdload_settings_page import WHDLoadSettingsPage
from launcher.ui.pageddialog import PagedDialog

SPACE = ""


class SettingsDialog(PagedDialog):
    @classmethod
    def open(cls, parent=None):
        return fsui.open_window_instance(cls, parent)

    def __init__(self, parent, index=0):
        PagedDialog.__init__(
            self,
            parent,
            "{} - {} Launcher".format(gettext("Settings"), fsgs.product),
        )

        # FIXME: remove this once the dialog uses Window as base class
        # self.setAttribute(Qt.WA_DeleteOnClose, True)

        # self.add_page(
        #     # gettext("Appearance"), LanguageSettingsPage,
        #     gettext("Language"), LanguageSettingsPage,
        #     fsui.Icon("language-settings", "pkg:workspace"))
        self.add_page(
            gettext("Common"),
            LanguageSettingsPage,
            fsui.Icon("language-settings", "pkg:workspace"),
            bold=True,
        )
        self.add_page(
            gettext("Controllers"),
            JoystickSettingsPage,
            fsui.Icon("gamepad", "pkg:workspace"),
        )
        self.add_page(
            gettext("Plugins"),
            PluginsSettingsPage,
            fsui.Icon("settings", "pkg:workspace"),
        )
        # self.add_page(
        #     gettext("Directories"),
        #     DirectoriesSettingsPage,
        #     fsui.Icon("folder", "pkg:launcher"),
        # )
        self.add_page(
            gettext("Advanced"),
            AdvancedSettingsPage,
            fsui.Icon("settings", "pkg:workspace"),
        )
        self.add_page(
            "FS-UAE",
            FSUAESettingsPage,
            fsui.Icon("fs-uae", "pkg:launcher"),
            bold=True,
        )
        self.add_page(
            gettext("Keyboard"),
            KeyboardSettingsPage,
            fsui.Icon("keyboard-settings", "pkg:workspace"),
        )
        self.add_page(
            gettext("Mouse"),
            MouseSettingsPage,
            fsui.Icon("mouse-settings", "pkg:workspace"),
        )
        self.add_page(
            gettext("Audio"),
            AudioSettingsPage,
            fsui.Icon("audio-settings", "pkg:workspace"),
        )
        self.add_page(
            gettext("Video"),
            VideoSettingsPage,
            fsui.Icon("video-settings", "pkg:workspace"),
        )
        self.add_page(
            gettext("Advanced Video"),
            AdvancedVideoSettingsPage,
            fsui.Icon("video-settings", "pkg:workspace"),
        )
        # self.add_page(
        #     gettext("Synchronization"), VideoSyncSettingsPage,
        #     fsui.Icon("video-settings", "pkg:workspace"))
        # self.add_page(
        #     gettext("Filters & Scaling"), FilterSettingsPage,
        #     fsui.Icon("video-settings", "pkg:workspace"))
        # self.add_page(gettext("OpenGL Settings"), OpenGLSettingsPage)
        # if Settings.get("database_feature") == "1":
        # self.add_page(
        #     gettext("Logging"), LoggingSettingsPage,
        #     fsui.Icon("settings", "pkg:workspace"))
        self.add_page(
            "{} Launcher".format(fsgs.product),
            LauncherSettingsPage,
            fsui.Icon("fs-uae-launcher", "pkg:launcher"),
            bold=True,
        )
        self.add_page(
            gettext("File Database"),
            ScanSettingsPage,
            fsui.Icon("indexing-settings", "pkg:workspace"),
        )
        self.add_page(
            gettext("Game Database"),
            GameDatabaseSettingsPage,
            fsui.Icon("database-settings", "pkg:workspace"),
        )
        if fsgs.openretro or settings.get(Option.PLATFORMS_FEATURE) == "1":
            self.add_page(
                gettext("Game Platforms"),
                GamePlatformsSettingsPage,
                fsui.Icon("database-settings", "pkg:workspace"),
            )
        # self.add_page(gettext("Custom Settings"), CustomSettingsPage)
        if LauncherSettings.get(Option.NETPLAY_FEATURE) != "0":
            self.add_page(
                gettext("Net Play"),
                NetplaySettingsPage,
                fsui.Icon("netplay-settings", "pkg:workspace"),
            )
        self.add_page(
            "WHDLoad", WHDLoadSettingsPage, fsui.Icon("hd", "pkg:launcher")
        )
        # self.add_page(
        #     gettext("Experimental Features"), ExperimentalFeaturesPage,
        #     fsui.Icon("settings", "pkg:workspace"))
        # self.add_page(
        #     gettext("Maintenance"), MaintenanceSettingsPage,
        #     fsui.Icon("maintenance", "pkg:workspace"))
        self.add_page(
            "{} Arcade".format(fsgs.product),
            ArcadeSettingsPage,
            fsui.Icon("fs-uae-arcade", "pkg:launcher"),
            bold=True,
        )

        # Old texts
        # gettext("Video Synchronization")
        # gettext("Synchronization")
        gettext("Advanced")

        last_index = self.get_page_index_by_title(
            LauncherSettings.get("last_settings_page")
        )
        index = last_index or index
        self.list_view.set_index(index)

        defaults_button = fsui.Button(self, gettext("Reset to Defaults"))
        defaults_button.activated.connect(self.__defaults_activated)
        self.button_layout.insert(0, defaults_button, fill=True)

        defaults_label = fsui.Label(
            self, gettext("Choices marked with (*) is the default setting")
        )
        self.button_layout.insert(1, defaults_label, margin_left=20)

        self.set_size((940, 560))
        # self.center_on_parent()

        self.closed.connect(self.__closed)
        self.page_changed.connect(self.__page_changed)

    def __page_changed(self):
        index = self.get_index()
        LauncherSettings.set("last_settings_page", self.get_page_title(index))

    def __closed(self):
        LauncherSignal.broadcast("settings_updated")

    def __defaults_activated(self):
        self.page.reset_to_defaults()



import fsui
from launcher.i18n import gettext
from launcher.launcher_settings import LauncherSettings
from launcher.ui.IconButton import IconButton
from fsbc.util import unused
from fsgs.FSGSDirectories import FSGSDirectories


class ScanPathsGroup(fsui.Group):
    def __init__(self, parent):
        fsui.Group.__init__(self, parent)
        self.layout = fsui.HorizontalLayout()

        self.layout2 = fsui.VerticalLayout()
        self.layout.add(self.layout2, fill=True, expand=True)

        layout3 = fsui.HorizontalLayout()
        self.layout2.add(layout3, fill=True, expand=True)

        self.list_view = fsui.ListView(self)
        self.list_view.set_min_height(130)
        self.default_icon = fsui.Image("launcher:res/folder_16.png")
        layout3.add(self.list_view, expand=True, fill=True)
        layout3.add_spacer(10)

        vlayout = fsui.VerticalLayout()
        layout3.add(vlayout, fill=True)

        add_button = IconButton(self, "add_button.png")
        add_button.set_tooltip(gettext("Add Directory to Search Path"))
        # add_button.disable()
        add_button.activated.connect(self.on_add_button)
        vlayout.add(add_button)
        vlayout.add_spacer(10)

        self.remove_button = IconButton(self, "remove_button.png")
        self.remove_button.set_tooltip(
            gettext("Remove Directory from Search Path")
        )
        self.remove_button.disable()
        self.remove_button.activated.connect(self.on_remove_button)
        vlayout.add(self.remove_button)

        # self.list_view.set_items(self.get_search_path())
        self.repopulate_list()
        self.list_view.on_select_item = self.on_select_item
        LauncherSettings.add_listener(self)

    def on_destroy(self):
        LauncherSettings.remove_listener(self)

    def on_setting(self, key, value):
        unused(value)
        if key == "search_path":
            self.repopulate_list()

    def on_select_item(self, index):
        unused(index)
        self.remove_button.enable()

    def repopulate_list(self):
        self.list_view.clear()
        for item in self.get_search_path():
            self.list_view.add_item(item, self.default_icon)

    @classmethod
    def get_search_path(cls):
        paths = FSGSDirectories.get_default_search_path()
        search_path = LauncherSettings.get("search_path")
        for p in search_path.split(";"):
            p = p.strip()
            if not p:
                continue
            elif p[0] == "-":
                p = p[1:]
                if p in paths:
                    paths.remove(p)
            else:
                if p not in paths:
                    paths.append(p)
        # The Configurations dir is always scanned on startup (whenever
        # modification time has changed). If we don't include it here too
        # always, the result will be that the configs sometimes appear (on
        # startup) and disappear (on scan).
        if not FSGSDirectories.get_configurations_dir() in paths:
            paths.append(FSGSDirectories.get_configurations_dir())
        # Likewise, we force the Kickstarts directory to always be scanned.
        if not FSGSDirectories.get_kickstarts_dir() in paths:
            paths.append(FSGSDirectories.get_kickstarts_dir())
        return paths

    def on_add_button(self):
        search_path = LauncherSettings.get("search_path")
        search_path = [x.strip() for x in search_path.split(";") if x.strip()]
        path = fsui.pick_directory(parent=self.get_window())
        if path:
            for i in range(len(search_path)):
                if search_path[i].startswith("-"):
                    if path == search_path[i][1:]:
                        search_path.remove(search_path[i])
                        break
                else:
                    if search_path[i] == path:
                        # Already added.
                        break
            else:
                default_paths = FSGSDirectories.get_default_search_path()
                if path not in default_paths:
                    search_path.append(path)
            LauncherSettings.set("search_path", ";".join(search_path))

    def on_remove_button(self):
        path = self.list_view.get_item(self.list_view.get_index())
        search_path = LauncherSettings.get("search_path")
        search_path = [x.strip() for x in search_path.split(";") if x.strip()]
        for i in range(len(search_path)):
            if search_path[i].startswith("-"):
                if path == search_path[i][1:]:
                    # Already removed.
                    break
            else:
                if search_path[i] == path:
                    search_path.remove(search_path[i])
                    break
        default_paths = FSGSDirectories.get_default_search_path()
        if path in default_paths:
            search_path.append("-" + path)
        LauncherSettings.set("search_path", ";".join(search_path))



import threading
import traceback

from arcade.glui.opengl import gl
from fsui.qt import QGLWidget


# noinspection PyPep8Naming
class GLWidget(QGLWidget):
    def __init__(self, parent, callbacks):
        QGLWidget.__init__(self, parent)
        self._callbacks = callbacks
        self._initialized = False
        self._first_initialize_gl_call = True
        self._first_resize_gl_call = True
        self._first_paint_gl_call = True

        # set_black_background(self)
        # palette = self.palette()
        # palette.setColor(self.backgroundRole(), Qt.blue)
        # self.setPalette(palette)
        # self.setAutoFillBackground(True)
        # self.setStyleSheet("background-color: black;")

    def initializeGL(self):
        print("[OPENGL] GLWidget.initializeGL")
        if self._first_initialize_gl_call:
            print("[OPENGL]", threading.current_thread())
            self._first_initialize_gl_call = False
        try:
            # Moved most of the initialization to (first invocation of)
            # resizeGL due to issues on OS X.
            context = self.context()
            print("[OPENGL] Qt OpenGL information:")
            print("[OPENGL] Valid context:", context.isValid())
            gl_format = context.format()
            print("[OPENGL] Version major:", gl_format.majorVersion())
            print("[OPENGL] Version minor:", gl_format.minorVersion())
            print("[OPENGL] Context profile:", gl_format.profile())
            print("[OPENGL] Direct rendering:", gl_format.directRendering())
            print("[OPENGL] Depth buffer size:", gl_format.depthBufferSize())
            print("[OPENGL] Double buffering:", self.doubleBuffer())
            print("[OPENGL] Auto buffer swap:", self.autoBufferSwap())
        except Exception:
            traceback.print_exc()

    def resizeGL(self, width, height):
        print("[OPENGL] GLWidget.resizeGL", width, height)
        if self._first_resize_gl_call:
            print("[OPENGL]", threading.current_thread())
            self._first_resize_gl_call = False

        gl.load()
        try:
            if not self._initialized:
                if width == 160 and height == 160:
                    print("[OPENGL] WARNING: OS X 160x160 workaround(?)")
                    # work around bug(?) on os x
                    return
            # Fail-safes to prevent a viewport of size 0 (prevents some
            # potential divide-by-zero nastiness).
            # if width == 0:
            #     width = 100
            # if height == 0:
            #     height = 100
            if width < 480:
                print("[OPENGL] Setting minimum width 480")
                width = 480
            if height < 270:
                print("[OPENGL] Setting minimum height 270")
                height = 270
            if not self._initialized:
                self._callbacks.initialize(width, height)
                self._initialized = True
                return
            self._callbacks.resize(width, height)
        except Exception:
            traceback.print_exc()
        gl.unload()

    def paintGL(self):
        if self._first_paint_gl_call:
            print("[OPENGL] GLWidget.paintGL")
            print("[OPENGL]", threading.current_thread())
            self._first_paint_gl_call = False
        gl.load()
        try:
            self._callbacks.render()
        except Exception:
            traceback.print_exc()
            # import sys
            # sys.exit(1)
        gl.unload()



import fsui
from launcher.i18n import gettext
from launcher.option import Option
from launcher.settings.settings_page import SettingsPage


class AudioSettingsPage(SettingsPage):
    def __init__(self, parent):
        super().__init__(parent)
        icon = fsui.Icon("audio-settings", "pkg:workspace")
        gettext("Audio Settings")
        title = gettext("Audio")
        subtitle = ""
        self.add_header(icon, title, subtitle)

        self.add_option("volume")
        self.add_option("stereo_separation")

        self.add_section(gettext("Floppy Drive Sound Emulation"))
        self.add_option("floppy_drive_volume")
        self.add_option(Option.FLOPPY_DRIVE_VOLUME_EMPTY)

        self.add_section(gettext("Advanced Audio Options"))
        self.add_option("audio_frequency")
        self.add_option("audio_buffer_target_size")



import os
import sys
from fsui.qt import QUrl, QLibraryInfo
from PyQt5.QtQuick import QQuickView

# to make sure cxFreeze includes it
import PyQt5.QtNetwork
import PyQt5.QtQml
from fsbc.application import app


class GameCenterView(QQuickView):
    def __init__(self, parent=None):
        QQuickView.__init__(self, parent)
        if getattr(sys, "frozen", ""):
            qml_path = os.path.join(app.executable_dir(), "qml")
        else:
            qml_path = os.path.expanduser("~/git/fs-uae/fs-uae-launcher/qml")

        engine = self.engine()
        print(engine.importPathList())
        print(engine.pluginPathList())
        # engine.setPluginPathList([qml_path, "."])

        # engine.addPluginPath(qml_path)
        # engine.addImportPath(qml_path)

        # engine.setPluginPathList([qml_path, "."])
        # engine.setImportPathList([qml_path])

        # engine.addPluginPath(qml_path)
        # print("setImportPathList", [QUrl.fromLocalFile(qml_path)])
        self.setSource(
            QUrl.fromLocalFile(
                os.path.join(qml_path, "ScaledUserInterface.qml")
            )
        )

        #     "qml/ScaledUserInterface.qml")))
        # self.setSource(QUrl.fromLocalFile(
        # self.setSource(QUrl.fromLocalFile(
        #     os.path.expanduser("~/git/fs-uae/fs-uae-launcher/qml/ScaledUserInterface.qml")))
        # self.setSource(QUrl.fromLocalFile(
        #     os.path.expanduser("~/git/fs-uae/fs-uae-launcher/qml/ScaledUserInterface.qml")))
        # print(QLibraryInfo.location(QLibraryInfo.Qml2ImportsPath))



import threading

from arcade.glui.opengl import gl
from arcade.glui.render import Render
from arcade.glui.texture import Texture

CACHE_SIZE = 50
# QT stores RGBA data in BGRA data (on little-endian platforms at least)
TEXTURE_FORMAT = gl.GL_BGRA


class TextureManager(object):
    _instance = None  # type: TextureManager

    @classmethod
    def get(cls):
        if not cls._instance:
            cls._instance = cls()
        return cls._instance

    @classmethod
    def reset(cls):
        cls._instance = None

    def __init__(self):
        self.lock = threading.Lock()
        self.image_list = []
        self.delay_set = set()
        self.image_dict = {}
        self.texture_dict = {}
        self.transparent = "\x00" * 512 * 512 * 4

    def set_image(self, image_path, image, size):
        with self.lock:
            if image_path not in self.image_dict:
                # no longer interested
                return
            self.image_dict[image_path] = (image, size)

    def load_images(self, image_paths):
        with self.lock:
            for ip in reversed(image_paths):
                if ip in self.image_dict:
                    self.image_list.remove(ip)
                    # self.image_list.insert(0, ip)
                else:
                    # texture = False means not loaded
                    # texture = None means failure loading
                    self.image_dict[ip] = False, None
                    self.texture_dict[ip] = None
                self.image_list.insert(0, ip)

            for ip in self.image_list[CACHE_SIZE:]:
                texture = self.texture_dict[ip]
                if texture is not None:
                    if texture is not Texture.default_item:
                        Render.get().delete_texture_list.append(
                            texture.texture
                        )
                del self.image_dict[ip]
                del self.texture_dict[ip]
            self.image_list = self.image_list[:CACHE_SIZE]

    def get_texture(self, ip):
        if ip is None:
            return None
        with self.lock:
            texture = self.texture_dict[ip]
            if texture:
                if ip in self.delay_set:
                    # do not return the texture the first time we ask for it,
                    # this is to allow the GPU driver time to upload the
                    # texture in the background (DMA transfer) without
                    # blocking for it
                    self.delay_set.remove(ip)
                    return None
            return texture

    def load_textures(self, max_num=10):
        loaded = 0
        target = gl.GL_TEXTURE_2D
        with self.lock:
            try:
                for ip in self.image_list:
                    if loaded == max_num:
                        break
                    texture = self.texture_dict[ip]
                    if texture is not None:
                        continue
                    pixels, size = self.image_dict[ip]
                    if pixels is False:
                        # queued
                        # self.texture_dict[ip] = Texture.default_item
                        continue
                    if pixels is None:
                        # could not load
                        # self.texture_dict[ip] = Texture(0, size=(0, 0))
                        # self.texture_dict[ip] = Texture.default_item
                        self.texture_dict[ip] = Texture.default_item
                        continue

                    texture = gl.glGenTextures(1)
                    self.texture_dict[ip] = Texture(
                        texture, size=(size[0], size[1])
                    )
                    self.delay_set.add(ip)

                    self.lock.release()

                    gl.glBindTexture(target, texture)
                    gl.glTexImage2D(
                        target,
                        0,
                        gl.GL_RGBA,
                        size[0],
                        size[1],
                        0,
                        TEXTURE_FORMAT,
                        gl.GL_UNSIGNED_BYTE,
                        pixels,
                    )
                    gl.glTexParameteri(
                        target, gl.GL_TEXTURE_MIN_FILTER, gl.GL_LINEAR
                    )
                    gl.glTexParameteri(
                        target, gl.GL_TEXTURE_MAG_FILTER, gl.GL_LINEAR
                    )
                    gl.glTexParameteri(
                        target, gl.GL_TEXTURE_WRAP_S, gl.GL_CLAMP_TO_EDGE
                    )
                    gl.glTexParameteri(
                        target, gl.GL_TEXTURE_WRAP_T, gl.GL_CLAMP_TO_EDGE
                    )
                    loaded += 1

                    self.lock.acquire()
            except Exception as e:
                print(e)

    def unload_textures(self):
        print("TextureManager.unload_textures")
        with self.lock:
            for ip in self.image_list:
                texture = self.texture_dict[ip]
                if texture is not None:
                    if texture is not Texture.default_item:
                        Render.get().delete_texture_list.append(
                            texture.texture
                        )
                self.texture_dict[ip] = None



import os
from pkg_resources import resource_filename, resource_stream

from .application import Application


class Resources(object):
    def __init__(self, package, subdir=""):
        self.package = package
        self.subdir = subdir
        self.resource_sha1s = {}

    def resource_name(self, resource):
        if self.subdir:
            # return os.path.join(self.subdir, resource)
            return self.subdir + "/" + resource
        return resource

    def stream(self, resource):
        print("Resources.stream", resource)
        resource_name = self.resource_name(resource)
        # return resource_stream(self.package, str_path(resource_name))
        try:
            return resource_stream(self.package, resource_name)
        except Exception as e:
            print(e)
        try:
            return open(self.path(resource), "rb")
        except FileNotFoundError:
            raise LookupError(resource)

    def path(self, resource):
        application = Application.get_instance()
        resource_name = self.resource_name(resource)
        print("looking up resource", resource_name)
        print("- package:", self.package)

        relative_path = os.path.join(self.package, resource_name)
        try:
            return application.data_file(relative_path)
        except LookupError:
            pass

        try:
            # print("resource_filename(\"{0}\", \"{1}\")".format(
            #     self.package, resource_name))
            return resource_filename(self.package, resource_name)
            # return resource_filename(
            #     self.package, Paths.encode(resource_name))
        except Exception:
            pass

        raise LookupError("Cannot find resource {0}".format(repr(resource)))



import os

import fsui
from fsgs.checksumtool import ChecksumTool
from fsgs.FSGSDirectories import FSGSDirectories
from fsgs.amiga import whdload
from launcher.i18n import gettext
from launcher.launcher_config import LauncherConfig
from launcher.ui.IconButton import IconButton
from launcher.ui.LauncherFilePicker import LauncherFilePicker
from launcher.ui.behaviors.platformbehavior import AmigaEnableBehavior


class HardDriveGroup(fsui.Panel):
    def __init__(self, parent, index):
        fsui.Panel.__init__(self, parent)
        AmigaEnableBehavior(self)
        self.layout = fsui.VerticalLayout()

        self.index = index
        self.config_key = "hard_drive_{0}".format(index)
        self.config_key_sha1 = "x_hard_drive_{0}_sha1".format(index)

        # if index == 0:
        #     # heading_label = fsui.HeadingLabel(self,
        #     #         _("Hard Drive {0}").format(index + 1))
        #     heading_label = fsui.HeadingLabel(self, gettext("Hard Drives"))
        #     self.layout.add(heading_label, margin_bottom=20)
        #     self.layout.add_spacer(0)

        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)

        self.eject_button = IconButton(self, "eject_button.png")
        self.eject_button.set_tooltip(gettext("Eject"))
        self.eject_button.activated.connect(self.on_eject_button)
        hori_layout.add(self.eject_button)

        self.text_field = fsui.TextField(self, "", read_only=True)
        hori_layout.add(self.text_field, expand=True, margin_left=10)

        self.browse_button = IconButton(self, "browse_folder_16.png")
        self.browse_button.set_tooltip(gettext("Browse for Folder"))
        self.browse_button.activated.connect(self.on_browse_folder_button)
        hori_layout.add(self.browse_button, margin_left=10)

        self.browse_button = IconButton(self, "browse_file_16.png")
        self.browse_button.set_tooltip(gettext("Browse for File"))
        self.browse_button.activated.connect(self.on_browse_file_button)
        hori_layout.add(self.browse_button, margin_left=10)

        self.initialize_from_config()
        self.set_config_handlers()

    def initialize_from_config(self):
        self.on_config(self.config_key, LauncherConfig.get(self.config_key))

    def set_config_handlers(self):
        LauncherConfig.add_listener(self)

    def on_destroy(self):
        LauncherConfig.remove_listener(self)

    def on_config(self, key, value):
        if key != self.config_key:
            return
        dir_path, name = os.path.split(value)
        if dir_path:
            path = "{0} ({1})".format(name, dir_path)
        else:
            path = name
        self.text_field.set_text(path)
        self.text_field.set_cursor_position(0)
        self.eject_button.enable(bool(value))

    def on_eject_button(self):
        LauncherConfig.set_multiple(
            [(self.config_key, ""), (self.config_key_sha1, "")]
        )

    def on_browse_folder_button(self):
        self.browse(dir_mode=True)

    def on_browse_file_button(self):
        self.browse(dir_mode=False)

    def browse(self, dir_mode):
        default_dir = FSGSDirectories.get_hard_drives_dir()
        dialog = LauncherFilePicker(
            self.get_window(),
            gettext("Choose Hard Drive"),
            "hd",
            LauncherConfig.get(self.config_key),
            dir_mode=dir_mode,
        )
        if not dialog.show_modal():
            dialog.destroy()
            return
        path = dialog.get_path()
        dialog.destroy()

        checksum_tool = ChecksumTool(self.get_window())
        sha1 = ""
        if dir_mode:
            print("not calculating HD checksums for directories")
        else:
            size = os.path.getsize(path)
            if size < 64 * 1024 * 1024:
                sha1 = checksum_tool.checksum(path)
            else:
                print("not calculating HD checksums HD files > 64MB")
        full_path = path

        # FIXME: use contract function
        dir_path, file = os.path.split(path)
        self.text_field.set_text(file)
        if os.path.normcase(os.path.normpath(dir_path)) == os.path.normcase(
            os.path.normpath(default_dir)
        ):
            path = file

        self.text_field.set_text(path)
        values = [(self.config_key, path), (self.config_key_sha1, sha1)]
        if self.index == 0:
            # whdload_args = ""
            # dummy, ext = os.path.splitext(path)
            # if not dir_mode and ext.lower() in Archive.extensions:
            #     try:
            #         whdload_args = self.calculate_whdload_args(full_path)
            #     except Exception:
            #         traceback.print_exc()
            # values.append(("x_whdload_args", whdload_args))
            values.extend(
                whdload.generate_config_for_archive(
                    full_path, model_config=False
                ).items()
            )
        LauncherConfig.set_multiple(values)



from __future__ import print_function

import os
import struct

from RomAccess import RomAccess

class KickRomAccess(RomAccess):

  EXT_HEADER_SIZE = 0x10
  FOOTER_SIZE = 0x18
  ROMHDR_SIZE = 8
  ROMHDR_256K = 0x11114ef9
  ROMHDR_512K = 0x11144ef9

  def __init__(self, rom_data):
    RomAccess.__init__(self, rom_data)

  def is_kick_rom(self):
    return self.detect_kick_rom() == "ok"

  def detect_kick_rom(self):
    if not self.check_size():
      return "size check failed"
    if not self.check_header():
      return "header check failed"
    if not self.check_footer():
      return "footer check failed"
    if not self.check_rom_size_field():
      return "rom size field mismatch"
    if not self.verify_check_sum():
      return "check sum mismatch"
    return "ok"

  def check_header(self):
    # expect 0x1114 0x4ef9
    val = self.read_long(0)
    if self.kib == 512:
      return val == self.ROMHDR_512K
    elif self.kib == 256:
      return val == self.ROMHDR_256K
    else:
      return False

  def check_kickety_split(self):
    # expect 0x1111 0x4ef9
    if self.kib == 512:
      val = self.read_long(0x40000)
      return val == self.ROMHDR_256K
    else:
      return False

  def check_footer(self):
    # expect 0x0019 ... 0x001f
    off = self.size - 14
    num = 0x19
    for i in xrange(7):
      val = self.read_word(off)
      if val != num:
        return False
      num += 1
      off += 2
    return True

  def check_size(self):
    # expect 256k or 512k ROM
    if self.size % 1024 != 0:
      return False
    if self.kib not in (256, 512):
      return False
    return True

  def check_rom_size_field(self):
    return self.read_rom_size_field() == self.size

  def check_magic_reset(self):
    return self.read_word(0xd0) == 0x4e70

  def calc_check_sum(self, skip_off=None):
    """Check internal kickstart checksum and return True if is correct"""
    chk_sum = 0
    num_longs = self.size / 4
    off = 0
    max_u32 = 0xffffffff
    for i in xrange(num_longs):
      val = struct.unpack_from(">I", self.rom_data, off)[0]
      if off != skip_off:
        chk_sum += val
      off += 4
      if chk_sum > max_u32:
        chk_sum = chk_sum & max_u32
        chk_sum += 1
    return max_u32 - chk_sum

  def verify_check_sum(self):
    chk_sum = self.calc_check_sum()
    return chk_sum == 0

  def read_check_sum(self):
    sum_off = self.size - 0x18
    return self.read_long(sum_off)

  def recalc_check_sum(self):
    sum_off = self.size - 0x18
    return self.calc_check_sum(sum_off)

  def write_check_sum(self):
    cs = self.recalc_check_sum()
    sum_off = self.size - 0x18
    self.write_long(sum_off, cs)
    return cs

  def write_rom_size_field(self):
    off = self.size - 0x14
    self.write_long(off, self.size)

  def write_header(self, jump_addr, kickety_split=False):
    if kickety_split:
      offset = 0x40000
      hdr = self.ROMHDR_256K
    elif self.kib == 256:
      offset = 0
      hdr = self.ROMHDR_256K
    else:
      offset = 0
      hdr = self.ROMHDR_512K
    self.write_long(offset, hdr)
    self.write_long(offset+4, jump_addr)

  def write_ext_header(self, jump_addr, rom_rev):
    self.write_header(jump_addr)
    self.write_word(8,0)
    self.write_word(10,0xffff)
    self.write_word(12,rom_rev[0])
    self.write_word(14,rom_rev[1])

  def write_ext_footer(self):
    self.write_footer()
    self.write_rom_size_field()
    self.write_check_sum()

  def write_footer(self):
    off = self.size - 0x10
    num = 0x18
    for i in xrange(8):
      self.write_word(off, num)
      num += 1
      off += 2

  def write_rom_ver_rev(self, rom_rev):
    """get (ver, rev) version info from ROM"""
    return struct.pack_into(">HH", self.rom_data, 12, rom_rev[0], rom_rev[1])

  def read_boot_pc(self):
    """return PC for booting the ROM"""
    return self.read_long(4)

  def read_rom_ver_rev(self):
    """get (ver, rev) version info from ROM"""
    return struct.unpack_from(">HH", self.rom_data, 12)

  def read_exec_ver_rev(self):
    """get (ver, rev) version info from ROM"""
    return struct.unpack_from(">HH", self.rom_data, 16)

  def read_rom_size_field(self):
    """return size of ROM stored in ROM itself"""
    off = self.size - 0x14
    return self.read_long(off)

  def get_base_addr(self):
    return self.read_boot_pc() & ~0xffff


class Loader(object):
  """Load kick rom images in different formats"""
  @classmethod
  def load(cls, kick_file, rom_key_file=None):
    raw_img = None
    rom_key = None
    # read rom image
    with open(kick_file, "rb") as fh:
      raw_img = fh.read()
    # coded rom?
    need_key = False
    if raw_img[:11] == 'AMIROMTYPE1':
      rom_img = raw_img[11:]
      need_key = True
    else:
      rom_img = raw_img
    # decode rom
    if need_key:
      # read key file
      if rom_key_file is None:
        path = os.path.dirname(kick_file)
        rom_key_file = os.path.join(path, "rom.key")
      with open(rom_key_file, "rb") as fh:
        rom_key = fh.read()
      rom_img = cls._decode(rom_img, rom_key)
    return rom_img

  @classmethod
  def _decode(cls, img, rom_key):
    data = bytearray(img)
    n = len(rom_key)
    for i in xrange(len(data)):
      off = i % n
      data[i] = data[i] ^ ord(rom_key[off])
    return bytes(data)


# tiny test
if __name__ == '__main__':
  import sys
  args = sys.argv
  n = len(args)
  if n > 1:
    ks_file = args[1]
  else:
    ks_file = 'amiga-os-310-a500.rom'
  print(ks_file)
  ks = Loader.load(ks_file,'rom.key')
  kh = KickRomAccess(ks)
  print("is_kick_rom", kh.is_kick_rom())
  print("detect_kick_rom", kh.detect_kick_rom())
  print("pc=%08x" % kh.read_boot_pc())
  print("ver,rev=", kh.read_rom_ver_rev())
  print("ver,rev=", kh.read_exec_ver_rev())
  print("size %08x == %08x" % (kh.read_rom_size_field(), len(ks)))
  print("base %08x" % kh.get_base_addr())
  print("get chk_sum=%08x" % kh.read_check_sum())
  print("calc chk_sum=%08x" % kh.recalc_check_sum())
#  with open("out.rom", "wb") as fh:
#    fh.write(ks.get_data())




import fsui
from launcher.ui.skin import Skin
from launcher.ui.config.InputGroup import InputGroup


class InputPanel(fsui.Panel):
    def __init__(self, parent):
        fsui.Panel.__init__(self, parent)
        Skin.set_background_color(self)
        self.layout = fsui.VerticalLayout()
        self.input_group = InputGroup(self, refresh_button=True)
        self.layout.add(self.input_group, fill=True)
        self.layout.add_spacer(Skin.EXTRA_GROUP_MARGIN)
        self.input_group = InputGroup(self, parallel_ports=True)
        self.layout.add(self.input_group, fill=True)
        self.input_group = InputGroup(self, custom_ports=True)
        self.layout.add(self.input_group, fill=True)



import hashlib
import os
import traceback
from configparser import ConfigParser, NoSectionError

from fsbc.paths import Paths
from fsbc.signal import Signal
from fsgs.checksumtool import ChecksumTool
from fsgs.FSGSDirectories import FSGSDirectories
from fsgs.amiga.amiga import Amiga
from fsgs.amiga.valueconfigloader import ValueConfigLoader
from fsgs.context import fsgs
from fsgs.platform import PlatformHandler
from launcher.option import Option
from .launcher_settings import LauncherSettings

cfg = [
    ("amiga_model", "", "checksum", "sync"),
    ("ntsc_mode", "", "checksum", "sync"),
    ("accuracy", "", "checksum", "sync"),
    ("chip_memory", "", "checksum", "sync"),
    ("slow_memory", "", "checksum", "sync"),
    ("fast_memory", "", "checksum", "sync"),
    ("zorro_iii_memory", "", "checksum", "sync"),
    ("bsdsocket_library", "", "checksum", "sync"),
    ("uaegfx_card", "", "checksum", "sync"),
    ("joystick_port_0", ""),
    ("joystick_port_0_mode", "", "checksum", "sync"),
    ("joystick_port_0_autofire", "", "checksum", "sync"),
    ("joystick_port_1", ""),
    ("joystick_port_1_mode", "", "checksum", "sync"),
    ("joystick_port_1_autofire", "", "checksum", "sync"),
    ("joystick_port_2", ""),
    ("joystick_port_2_mode", "", "checksum", "sync"),
    ("joystick_port_2_autofire", "", "checksum", "sync"),
    ("joystick_port_3", ""),
    ("joystick_port_3_mode", "", "checksum", "sync"),
    ("joystick_port_3_autofire", "", "checksum", "sync"),
    ("floppy_drive_count", "", "checksum", "sync"),
    ("cdrom_drive_count", "", "checksum", "sync"),
    # this is not an Amiga device, so no need to checksum / sync
    ("joystick_port_4_mode", "", "custom"),
    ("kickstart_file", ""),
    ("x_kickstart_file", "", "no_save"),
    ("x_kickstart_file_sha1", "", "checksum", "sync", "no_save"),
    ("kickstart_ext_file", ""),
    ("x_kickstart_ext_file", "", "no_save"),
    ("x_kickstart_ext_file_sha1", "", "checksum", "sync", "no_save"),
    (Option.X_WHDLOAD_ARGS, "", "checksum", "sync"),
    (Option.X_WHDLOAD_VERSION, "", "checksum", "sync"),
    # (Option.WHDLOAD_SPLASH_DELAY, "", "checksum", "sync"),
    # (Option.WHDLOAD_QUIT_KEY, "", "checksum", "sync"),
    ("floppy_drive_count", "", "checksum", "sync"),
    ("floppy_drive_speed", "", "checksum", "sync"),
    ("cdrom_drive_count", "", "checksum", "sync"),
    ("dongle_type", "", "checksum", "sync"),
    ("__netplay_game", "", "checksum", "sync"),
    ("__netplay_password", "", "checksum", "sync"),
    ("__netplay_players", "", "checksum", "sync"),
    ("__netplay_port", "", "sync"),
    ("__netplay_addresses", "", "checksum", "sync"),
    ("__netplay_host", ""),
    ("__netplay_ready", ""),
    ("__netplay_state_dir_name", "", "checksum", "sync"),
    ("__version", "FIXME"),
    ("__error", ""),
    ("x_game_uuid", ""),
    ("x_game_xml_path", ""),
    ("title", "", "custom"),
    ("sub_title", "", "custom"),
    ("viewport", "", "custom"),
    ("year", ""),
    ("developer", ""),
    ("publisher", ""),
    ("languages", ""),
    ("players", ""),
    ("protection", ""),
    ("hol_url", ""),
    ("wikipedia_url", ""),
    ("database_url", ""),
    ("lemon_url", ""),
    ("mobygames_url", ""),
    ("amigamemo_url", ""),
    ("whdload_url", ""),
    ("mobygames_url", ""),
    ("thelegacy_url", ""),
    ("homepage_url", ""),
    ("longplay_url", ""),
    ("variant_uuid", ""),
    ("game_uuid", ""),
    ("download_file", ""),
    ("download_page", ""),
    ("download_terms", ""),
    ("download_notice", ""),
    ("x_missing_files", ""),
    ("x_game_notice", ""),
    ("x_variant_notice", ""),
    ("x_variant_warning", ""),
    ("x_variant_error", ""),
    ("x_joy_emu_conflict", ""),
    ("screen1_sha1", ""),
    ("screen2_sha1", ""),
    ("screen3_sha1", ""),
    ("screen4_sha1", ""),
    ("screen5_sha1", ""),
    ("front_sha1", ""),
    ("title_sha1", ""),
    ("mouse_integration", "", "checksum", "sync"),
    ("cdrom_drive_0_delay", "", "checksum", "sync"),
    ("cpu", "", "checksum", "sync"),
    ("graphics_card", "", "checksum", "sync"),
    ("graphics_memory", "", "checksum", "sync"),
    ("accelerator", "", "checksum", "sync"),
    ("accelerator_memory", "", "checksum", "sync"),
    ("blizzard_scsi_kit", "", "checksum", "sync"),
    ("motherboard_ram", "", "checksum", "sync"),
    ("sound_card", "", "checksum", "sync"),
    ("jit_compiler", "", "checksum", "sync"),
    ("__database", ""),
    ("platform", ""),
    ("save_disk", "", "checksum", "sync"),
    ("network_card", "", "checksum", "sync"),
    ("freezer_cartridge", "", "checksum", "sync"),
    (Option.A7800_A78_HEADER, "", "checksum", "sync"),
    (Option.A7800_MODEL, "", "checksum", "sync"),
    (Option.A7800_PORT_1_TYPE, "", "checksum", "sync"),
    (Option.A7800_PORT_2_TYPE, "", "checksum", "sync"),
    (Option.NES_MODEL, "", "checksum", "sync"),
    (Option.ATARI_MODEL, "", "checksum", "sync"),
    (Option.C64_MODEL, "", "checksum", "sync"),
    (Option.CARTRIDGE_SLOT, "", "checksum", "sync"),
    (Option.FILE_LIST, "", "custom", "sync"),
    (Option.FLOPPY_DRIVE_VOLUME_EMPTY, "", "sync"),
    (Option.GAME_NAME, ""),
    ("mame_rom_set", "", "custom", "checksum", "sync"),
    (Option.NEOGEO_MODEL, "", "checksum", "sync"),
    (Option.NES_INES_HEADER, "", "checksum", "sync"),
    (Option.NES_MODEL, "", "checksum", "sync"),
    (Option.SMD_EMULATOR, "", "checksum", "sync"),
    (Option.SMD_MODEL, "", "checksum", "sync"),
    (Option.SMD_PORT_1_TYPE, "", "checksum", "sync"),
    (Option.SMD_PORT_2_TYPE, "", "checksum", "sync"),
    ("refresh_rate", "", "custom", "checksum", "sync"),
    (Option.TAPE_DRIVE_0, "", "checksum", "sync"),
    ("tape_image_0", "", "checksum", "sync"),
    (Option.VARIANT_NAME, ""),
    (Option.ZXS_MODEL, "", "checksum", "sync"),
]

for _i in range(Amiga.MAX_FLOPPY_DRIVES):
    cfg.append(("floppy_drive_{0}".format(_i), ""))
    cfg.append(
        (
            "x_floppy_drive_{0}_sha1".format(_i),
            "",
            "checksum",
            "sync",
            "no_save",
        )
    )
for _i in range(Amiga.MAX_FLOPPY_IMAGES):
    cfg.append(("floppy_image_{0}".format(_i), ""))
    cfg.append(
        (
            "x_floppy_image_{0}_sha1".format(_i),
            "",
            "checksum",
            "sync",
            "no_save",
        )
    )
for _i in range(Amiga.MAX_CDROM_DRIVES):
    cfg.append(("cdrom_drive_{0}".format(_i), ""))
    cfg.append(
        (
            "x_cdrom_drive_{0}_sha1".format(_i),
            "",
            "checksum",
            "sync",
            "no_save",
        )
    )
for _i in range(Amiga.MAX_CDROM_IMAGES):
    cfg.append(("cdrom_image_{0}".format(_i), ""))
    cfg.append(
        (
            "x_cdrom_image_{0}_sha1".format(_i),
            "",
            "checksum",
            "sync",
            "no_save",
        )
    )
for _i in range(Amiga.MAX_HARD_DRIVES):
    cfg.append(("hard_drive_{0}".format(_i), ""))
    cfg.append(
        ("hard_drive_{0}_label".format(_i), "", "checksum", "sync", "custom")
    )
    cfg.append(
        (
            "hard_drive_{0}_priority".format(_i),
            "",
            "checksum",
            "sync",
            "custom",
        )
    )
    cfg.append(
        ("x_hard_drive_{0}_sha1".format(_i), "", "checksum", "sync", "no_save")
    )


class LauncherConfig(object):
    config_keys = [x[0] for x in cfg]

    default_config = {}
    for c in cfg:
        default_config[c[0]] = c[1]

    key_order = [x[0] for x in cfg]
    checksum_keys = [x[0] for x in cfg if "checksum" in x]
    sync_keys_list = [x[0] for x in cfg if "sync" in x]
    sync_keys_set = set(sync_keys_list)
    no_custom_config = [x[0] for x in cfg if "custom" not in x]

    no_custom_config.append("__changed")
    no_custom_config.append("__ready")
    no_custom_config.append("__config_name")
    # no_custom_config.append("__database")
    # no_custom_config.append("x_whdload_icon")
    # no_custom_config.append("platform")

    no_save_keys_set = set([x[0] for x in cfg if "no_save" in x])

    reset_values = {}
    for i in range(Amiga.MAX_FLOPPY_DRIVES):
        reset_values["floppy_drive_{0}".format(i)] = (
            "x_floppy_drive_{0}_sha1".format(i),
            "",
        )
    for i in range(Amiga.MAX_FLOPPY_IMAGES):
        reset_values["floppy_image_{0}".format(i)] = (
            "x_floppy_image_{0}_sha1".format(i),
            "",
        )
    for i in range(Amiga.MAX_CDROM_DRIVES):
        reset_values["cdrom_drive_{0}".format(i)] = (
            "x_cdrom_drive_{0}_sha1".format(i),
            "",
        )
    for i in range(Amiga.MAX_CDROM_IMAGES):
        reset_values["cdrom_image_{0}".format(i)] = (
            "x_cdrom_image_{0}_sha1".format(i),
            "",
        )
    for i in range(Amiga.MAX_HARD_DRIVES):
        reset_values["hard_drive_{0}".format(i)] = (
            "x_hard_drive_{0}_sha1".format(i),
            "",
        )
    reset_values["x_kickstart_file"] = ("x_kickstart_file_sha1", "")
    reset_values["x_kickstart_ext_file"] = ("x_kickstart_ext_file_sha1", "")

    # config = default_config.copy()
    # config_listeners = []

    @classmethod
    def keys(cls):
        return fsgs.config.values.keys()

    @classmethod
    def copy(cls):
        return fsgs.config.copy()

    @classmethod
    def get(cls, key, default=""):
        return fsgs.config.get(key, default)

    @classmethod
    def add_listener(cls, listener):
        # deprecated
        Signal("fsgs:config").connect(getattr(listener, "on_config"))

    @classmethod
    def remove_listener(cls, listener):
        # deprecated
        Signal("fsgs:config").disconnect(getattr(listener, "on_config"))

    @classmethod
    def set(cls, key, value):
        fsgs.config.set(key, value)

    @classmethod
    def set_multiple(cls, items):
        fsgs.config.set(items)

    @classmethod
    def update_from_config_dict(cls, config_dict):
        changes = []
        for key, value in config_dict.items():
            if key in fsgs.config.values:
                if fsgs.config.values[key] != value:
                    changes.append((key, value))
            else:
                changes.append((key, value))
        cls.set_multiple(changes)

    @classmethod
    def sync_items(cls):
        for key, value in fsgs.config.values.items():
            if key in cls.sync_keys_set:
                yield key, value

    @classmethod
    def checksum(cls):
        return cls.checksum_config(fsgs.config.copy())

    @classmethod
    def checksum_config(cls, config):
        s = hashlib.sha1()
        for key in cls.checksum_keys:
            value = config[key]
            s.update(str(value).encode("UTF-8"))
        return s.hexdigest()

    @classmethod
    def update_kickstart_in_config_dict(cls, config_dict):
        print("update_kickstart_in_config")
        model = config_dict.setdefault(
            "amiga_model", cls.default_config["amiga_model"]
        )

        kickstart_file = config_dict.setdefault("kickstart_file", "")
        if kickstart_file:
            config_dict["x_kickstart_file"] = config_dict["kickstart_file"]
            if kickstart_file == "internal":
                config_dict["x_kickstart_file_sha1"] = Amiga.INTERNAL_ROM_SHA1
            else:
                # FIXME: set checksum
                pass
        else:
            checksums = Amiga.get_model_config(model)["kickstarts"]
            for checksum in checksums:
                path = fsgs.file.find_by_sha1(checksum)
                if path:
                    config_dict["x_kickstart_file"] = path
                    config_dict["x_kickstart_file_sha1"] = checksum
                    break
            else:
                print("WARNING: no suitable kickstart file found")
                config_dict["x_kickstart_file"] = ""
                config_dict["x_kickstart_file_sha1"] = Amiga.INTERNAL_ROM_SHA1

        if config_dict.setdefault("kickstart_ext_file", ""):
            config_dict["x_kickstart_ext_file"] = config_dict[
                "kickstart_ext_file"
            ]
            # FIXME: set checksum
        else:
            checksums = Amiga.get_model_config(model)["ext_roms"]
            if len(checksums) == 0:
                config_dict["x_kickstart_ext_file"] = ""
                config_dict["x_kickstart_ext_file_sha1"] = ""
            else:
                for checksum in checksums:
                    path = fsgs.file.find_by_sha1(checksum)
                    if path:
                        config_dict["x_kickstart_ext_file"] = path
                        config_dict["x_kickstart_ext_file_sha1"] = checksum
                        break
                else:
                    # print("WARNING: no suitable kickstart ext file found")
                    config_dict["x_kickstart_ext_file"] = ""
                    config_dict["x_kickstart_ext_file_sha1"] = ""
                    # Warnings.set("hardware", "kickstart_ext",
                    #              "No suitable extended kickstart found")
                    # FIXME: set sha1 and name x_options also

    @classmethod
    def update_kickstart(cls):
        cls.set_kickstart_from_model()

    @classmethod
    def set_kickstart_from_model(cls):
        print("set_kickstart_from_model")
        config_dict = fsgs.config.values.copy()
        cls.update_kickstart_in_config_dict(config_dict)
        cls.update_from_config_dict(config_dict)

    @classmethod
    def load_default_config(cls, platform=""):
        print("load_default_config platform =", platform)
        if platform:
            cls.load({Option.PLATFORM: platform})
        else:
            cls.load({})
        LauncherSettings.set(Option.CONFIG_NAME, "Unnamed Configuration")
        LauncherSettings.set("config_path", "")
        LauncherSettings.set("config_xml_path", "")

    @classmethod
    def load(cls, config):
        update_config = {}
        for key, value in cls.default_config.items():
            update_config[key] = value
        for key in list(fsgs.config.values.keys()):
            if key not in cls.default_config:
                # We need to broadcast changed for all config keys, also
                # the ones this class does not know about
                if key.startswith("__implicit_"):
                    # Unless it is an implicit config key, in which case
                    # it should be properly updated by another component, and
                    # we don't want to have the options flip-flop too much.
                    pass
                elif key in ["__changed"]:
                    # We handle this key explicitly, below.
                    pass
                else:
                    update_config[key] = ""

        for key, value in config.items():
            # if this is a settings key, change settings instead
            if key in LauncherSettings.initialize_from_config:
                LauncherSettings.set(key, value)
            else:
                update_config[key] = value

        cls.update_kickstart_in_config_dict(update_config)
        cls.fix_loaded_config(update_config)
        # print("about to set", update_config)
        cls.set_multiple(update_config.items())
        # Settings.set("config_changed", "0")
        if "__changed" in update_config:
            # We specifically loaded __changed, probably in fix_loaded_config.
            print("__changed was explicitly set:", update_config["__changed"])
        else:
            # Mark config as unchanged (i.e. does not need to be saved).
            cls.set("__changed", "0")

            # cls.update_kickstart()

    @classmethod
    def fix_joystick_ports(cls, config):
        # from .Settings import Settings

        print("---", config["joystick_port_0"])
        print("---", config["joystick_port_1"])

        from .devicemanager import DeviceManager

        available = DeviceManager.get_joystick_names()
        available.extend(["none", "mouse", "keyboard"])
        available_lower = [x.lower() for x in available]

        device_ids = DeviceManager.get_joystick_ids()
        device_ids.extend(["none", "mouse", "keyboard"])

        # remove devices from available list if specified and found
        try:
            index = available_lower.index(config["joystick_port_1"].lower())
        except ValueError:
            pass
        else:
            del available[index]
            del available_lower[index]
        try:
            index = available_lower.index(config["joystick_port_0"].lower())
        except ValueError:
            pass
        else:
            del available[index]
            del available_lower[index]

        # if config in
        # print("--------------------------------------------")
        if config["joystick_port_1_mode"] in ["joystick", "cd32 gamepad"]:
            if not config["joystick_port_1"]:
                want = LauncherSettings.get("primary_joystick").lower()
                # print("want", want)
                try:
                    index = available_lower.index(want)
                except ValueError:
                    index = -1
                print("available", available_lower)
                print("want", repr(want), "index", index)
                if index == -1:
                    index = len(available) - 1
                if index >= 0:
                    config["joystick_port_1"] = device_ids[index]
                    del available[index]
                    del available_lower[index]
                    del device_ids[index]
        elif config["joystick_port_1_mode"] in ["mouse"]:
            if not config["joystick_port_1"]:
                config["joystick_port_1"] = "mouse"
        elif config["joystick_port_1_mode"] in ["nothing"]:
            if not config["joystick_port_1"]:
                config["joystick_port_1"] = "none"

        if config["joystick_port_0_mode"] in ["joystick", "cd32 gamepad"]:
            if not config["joystick_port_0"]:
                want = LauncherSettings.get("secondary_joystick").lower()
                try:
                    index = available_lower.index(want)
                except ValueError:
                    index = -1
                # print("want", want, "index", index)
                if index == -1:
                    index = len(available) - 1
                if index >= 0:
                    config["joystick_port_0"] = device_ids[index]
                    del available[index]
                    del available_lower[index]
                    del device_ids[index]
        elif config["joystick_port_0_mode"] in ["mouse"]:
            if not config["joystick_port_0"]:
                config["joystick_port_0"] = "mouse"
        elif config["joystick_port_0_mode"] in ["nothing"]:
            if not config["joystick_port_0"]:
                config["joystick_port_0"] = "none"

    @classmethod
    def load_file(cls, path):
        try:
            return cls._load_file(path, "")
        except Exception:
            # FIXME: errors should be logged / displayed
            cls.load_default_config()
            traceback.print_exc()
        return False

    @classmethod
    def load_data(cls, data):
        print("Config.load_data")
        try:
            cls._load_file("", data)
        except Exception:
            # FIXME: errors should be logged / displayed
            cls.load_default_config()
            traceback.print_exc()

    @classmethod
    def create_fs_name(cls, name):
        name = name.replace(":", " - ")
        name = name.replace("*", "-")
        name = name.replace("?", "")
        name = name.replace("/", "-")
        name = name.replace("\\", "-")
        name = name.replace('"', "'")
        for i in range(3):
            name = name.replace("  ", " ")
        while name.endswith("."):
            name = name[:-1]
        name = name.strip()
        return name

    @classmethod
    def _load_file(cls, path, data):
        if data:
            print("loading config from data")
        else:
            print("loading config from " + repr(path))
            if not os.path.exists(path):
                print("config file does not exist")
        if data:
            raise Exception("_load_file (data) not implemented")
        else:
            config_xml_path = ""
            cp = ConfigParser(interpolation=None, strict=False)
            try:
                with open(path, "r", encoding="UTF-8") as f:
                    cp.read_file(f)
            except Exception as e:
                print(repr(e))
                return False
            config = {}
            try:
                keys = cp.options("config")
            except NoSectionError:
                keys = []
            for key in keys:
                config[key] = cp.get("config", key)
            try:
                keys = cp.options("fs-uae")
            except NoSectionError:
                keys = []
            for key in keys:
                config[key] = cp.get("fs-uae", key)

        LauncherSettings.set("config_path", path)

        print("__changed before load:", cls.get("__changed"))
        cls.load(config)
        print("__changed after load:", cls.get("__changed"))
        # Changed may have been set by fix_loaded_config
        changed = cls.get("__changed", "0")

        config_name = config.get("__config_name", "")
        if config_name:
            config_name = cls.create_fs_name(config_name)
        else:
            config_name, ext = os.path.splitext(os.path.basename(path))

        LauncherSettings.set(Option.CONFIG_NAME, config_name)
        LauncherSettings.set("config_xml_path", config_xml_path)
        cls.set("__changed", changed)
        return True

    @classmethod
    def load_values(cls, values, uuid=""):
        # print("loading config values", values)
        platform_id = values.get("platform", "").lower()

        if platform_id in ["amiga", "cdtv", "cd32"]:
            value_config_loader = ValueConfigLoader(uuid=uuid)
            value_config_loader.load_values(values)
            config = value_config_loader.get_config()
            cls.load(config)
            values["__config_name"] = config.get("__config_name")
        else:
            print("Warning: Non-Amiga game loaded")
            platform_handler = PlatformHandler.create(platform_id)
            loader = platform_handler.get_loader(fsgs)
            fsgs.config.load(loader.load_values(values))
        cls.post_load_values(values)

    @classmethod
    def post_load_values(cls, values):
        print("POST_LOAD_VALUES")
        if values.get("__config_name", ""):
            print("__config_name was set")
            config_name = values.get("__config_name", "")
        else:
            config_name = "{0} ({1})".format(
                values.get("game_name"), values.get("variant_name")
            )
        LauncherSettings.set("config_path", "")
        if config_name:
            config_name = cls.create_fs_name(config_name)
        LauncherSettings.set(Option.CONFIG_NAME, config_name)
        LauncherSettings.set("config_xml_path", "")
        cls.set("__changed", "0")

    @staticmethod
    def is_implicit_option(key):
        return key.startswith("__implicit_")

    @staticmethod
    def is_custom_uae_option(key):
        return key.startswith("uae_")

    @classmethod
    def is_custom_option(cls, key):
        return key not in cls.config_keys

    @classmethod
    def is_config_only_option(cls, key):
        if key in LauncherSettings.default_settings:
            # This key is specifically white-listed as a settings key
            return False
        if cls.is_custom_uae_option(key):
            return True
        if not cls.is_custom_option(key):
            return True
        return False

    @classmethod
    def fix_loaded_config(cls, config):
        print("[CONFIG] Fix loaded config")
        # cls.fix_joystick_ports(config)

        # FIXME: parent
        checksum_tool = ChecksumTool(None)

        def fix_file_checksum(sha1_key, key, base_dir, is_rom=False):
            path = config.get(key, "")
            # hack to synchronize URLs
            # print(repr(path))
            if path.startswith("http://") or path.startswith("https://"):
                sha1 = path
                config[sha1_key] = sha1
                return
            path = Paths.expand_path(path)
            sha1 = config.get(sha1_key, "")
            if not path:
                return
            if sha1:
                # assuming sha1 is correct
                return
            if not os.path.exists(path):
                print(repr(path), "does not exist")
                path = os.path.join(base_dir, path)
                if not os.path.exists(path):
                    print(repr(path), "does not exist")
                    return
            if os.path.isdir(path):
                # could set a fake checksum here or something, to indicate
                # that it isn't supposed to be set..
                return
            print("checksumming", repr(path))
            size = os.path.getsize(path)
            if size > 64 * 1024 * 1024:
                # not checksumming large files right now
                print("not checksumming large file")
                return

            if is_rom:
                sha1 = checksum_tool.checksum_rom(path)
            else:
                sha1 = checksum_tool.checksum(path)
            config[sha1_key] = sha1

        for i in range(Amiga.MAX_FLOPPY_DRIVES):
            fix_file_checksum(
                "x_floppy_drive_{0}_sha1".format(i),
                "floppy_drive_{0}".format(i),
                FSGSDirectories.get_floppies_dir(),
            )
        for i in range(Amiga.MAX_FLOPPY_IMAGES):
            fix_file_checksum(
                "x_floppy_image_{0}_sha1".format(i),
                "floppy_image_{0}".format(i),
                FSGSDirectories.get_floppies_dir(),
            )
        for i in range(Amiga.MAX_CDROM_DRIVES):
            fix_file_checksum(
                "x_cdrom_drive_{0}_sha1".format(i),
                "cdrom_drive_{0}".format(i),
                FSGSDirectories.get_cdroms_dir(),
            )
        for i in range(Amiga.MAX_CDROM_IMAGES):
            fix_file_checksum(
                "x_cdrom_image_{0}_sha1".format(i),
                "cdrom_image_{0}".format(i),
                FSGSDirectories.get_cdroms_dir(),
            )
        for i in range(Amiga.MAX_HARD_DRIVES):
            fix_file_checksum(
                "x_hard_drive_{0}_sha1".format(i),
                "hard_drive_{0}".format(i),
                FSGSDirectories.get_hard_drives_dir(),
            )

        # FIXME: need to handle checksums for Cloanto here
        fix_file_checksum(
            "x_kickstart_file_sha1",
            "x_kickstart_file",
            FSGSDirectories.get_kickstarts_dir(),
            is_rom=True,
        )
        fix_file_checksum(
            "x_kickstart_ext_file_sha1",
            "x_kickstart_ext_file",
            FSGSDirectories.get_kickstarts_dir(),
            is_rom=True,
        )

        # Convert uaegfx_card to new graphics_card option
        uae_gfx_card = config.get(Option.UAEGFX_CARD, "")
        if uae_gfx_card:
            if uae_gfx_card == "1":
                if not config.get(Option.GRAPHICS_CARD, ""):
                    config[Option.GRAPHICS_CARD] = "uaegfx"
            del config[Option.UAEGFX_CARD]
            # FIXME: Set changed!
            config["__changed"] = "1"



from fsui.qt import QUrl
from fsui.qt.helpers import QParent
from .widget import Widget
from PyQt5.QtWebKitWidgets import QWebView

# from PyQt5.QtWebEngineWidgets import QWebEngineView


class WebView(Widget):
    def __init__(self, parent):
        super().__init__()
        self.set_widget(QWebView(QParent(parent)))
        # self.widget().move(0, 2000)
        # if not border:
        #     self.setFrameStyle(fsui.qt.QFrame.NoFrame)
        # self.setReadOnly(read_only)
        # if font_family:
        #     print("FIXME: not respecting font_family yet")
        #     font = fsui.qt.QFont("Courier")
        #     # font.setStyleHint(QtGui.QFont.TypeWriter)
        #     self.setFont(font)
        # if text:
        #     self.append_text(text)
        # self.textChanged.connect(self.__text_changed)

    def load(self, url):
        print("load", url)
        self.widget().load(QUrl(url))



import fsui
from ..skin import Skin


BORDER_COLOR = (0xE5, 0xE5, 0xE5, 0xFF)


class BottomPanel(fsui.Panel):
    def __init__(self, parent):
        fsui.Panel.__init__(self, parent, paintable=True)
        Skin.set_background_color(self)

        self.color_2 = Skin.get_background_color()
        if self.color_2 is not None:
            self.color_1 = self.color_2.copy().darken(0.08)
            if fsui.System.macosx:
                self.line_color_1 = self.color_2.copy().mix(
                    fsui.Color(0x00, 0x00, 0x00), 0.25
                )
                self.line_color_2 = self.color_2.copy().mix(
                    fsui.Color(0xFF, 0xFF, 0xFF), 0.25
                )
            else:
                self.line_color_1 = self.color_2.copy().mix(
                    fsui.Color(0xFF, 0xFF, 0xFF)
                )
                self.line_color_2 = self.line_color_1
        else:
            # self.color_1 = fsui.Color(0x00, 0x00, 0x00, 0x20)
            # self.color_2 = fsui.Color(0x00, 0x00, 0x00, 0x00)
            self.color_1 = fsui.Color(0xFF, 0xFF, 0xFF, 0x60)
            self.color_2 = fsui.Color(0xFF, 0xFF, 0xFF, 0x00)
            self.line_color_1 = fsui.Color(0xFF, 0xFF, 0xFF, 0xA0)
            self.line_color_2 = self.line_color_1

        self.border_color = fsui.Color(*BORDER_COLOR)
        self.white_color = fsui.Color(0xFF, 0xFF, 0xFF, 0xFF)

    def on_paint(self):
        dc = self.create_dc()
        self.draw_background(dc)

    def draw_background(self, dc):
        size = self.size()
        if Skin.fws():
            dc.draw_rectangle(0, 0, size[0], 2, self.border_color)
            dc.draw_rectangle(0, 2, size[0], size[1] - 2, self.white_color)
            return
        dc.draw_line(0, 0, size[0], 0, self.line_color_1)
        dc.draw_line(0, 1, size[0], 1, self.line_color_2)
        dc.draw_vertical_gradient(
            0, 2, size[0], size[1] - 2, self.color_1, self.color_2
        )



import traceback

from .launcher_config import LauncherConfig


class ConfigChecker:
    def __init__(self):
        LauncherConfig.add_listener(self)

    def on_config(self, key, value):
        if key == "floppy_drive_0" and value:
            if not LauncherConfig.get("x_floppy_drive_0_sha1"):
                print("")
                traceback.print_stack()
                print("(floppy_drive_0 set without x_floppy_drive_0_sha1)")
                print("")

        elif key == "floppy_image_0" and value:
            if not LauncherConfig.get("x_floppy_image_0_sha1"):
                print("")
                traceback.print_stack()
                print("(floppy_image_0 set without x_floppy_image_0_sha1)")
                print("")

        elif key == "hard_drive_0" and value:
            if not LauncherConfig.get("x_hard_drive_0_sha1"):
                print("")
                traceback.print_stack()
                print("(hard_drive_0 set without x_hard_drive_0_sha1)")
                print("")

        elif key == "kickstart_file" and value:
            if not LauncherConfig.get("x_kickstart_file_sha1"):
                print("")
                traceback.print_stack()
                print("(kickstart_file set without x_kickstart_file_sha1)")
                print("")

        elif key == "x_kickstart_file" and value:
            if not LauncherConfig.get("x_kickstart_file_sha1"):
                print("")
                traceback.print_stack()
                print("(x_kickstart_file set without x_kickstart_file_sha1)")
                print("")

        elif key == "kickstart_ext_file" and value:
            if not LauncherConfig.get("x_kickstart_ext_file_sha1"):
                print("")
                traceback.print_stack()
                print(
                    "(kickstart_ext_file set without"
                    " x_kickstart_ext_file_sha1)"
                )
                print("")

        elif key == "x_kickstart_ext_file" and value:
            if not LauncherConfig.get("x_kickstart_ext_file_sha1"):
                print("")
                traceback.print_stack()
                print(
                    "(x_kickstart_ext_file set without"
                    " x_kickstart_ext_file_sha1)"
                )
                print("")



import weakref
from launcher.launcher_config import LauncherConfig

AMIGA_PLATFORMS = ["", "amiga", "cd32", "cdtv"]
FLOPPY_PLATFORMS = AMIGA_PLATFORMS + ["atari"]
CD_PLATFORMS = AMIGA_PLATFORMS + ["psx"]


class PlatformBehavior:
    def __init__(self, parent, platforms):
        self.platforms = set(platforms)
        parent.__platform_behavior = self
        self._parent = weakref.ref(parent)
        self.on_config("platform", LauncherConfig.get("platform"))
        LauncherConfig.add_listener(self)

        # FIXME: We need to disconnect the listener
        parent.destroyed.connect(self.on_destroy)

    def on_destroy(self, *_):
        LauncherConfig.remove_listener(self)

    def on_config(self, key, value):
        if key == "platform":
            platform = value.lower()
            self.perform(platform in self.platforms)

    def perform(self, match):
        pass


class PlatformEnableBehavior(PlatformBehavior):
    def __init__(self, parent, platforms):
        super().__init__(parent, platforms)

    def perform(self, match):
        widget = self._parent()
        if match:
            widget.enable()
        else:
            widget.disable()


class AmigaEnableBehavior(PlatformEnableBehavior):
    def __init__(self, parent):
        super().__init__(parent, platforms=AMIGA_PLATFORMS)


class FloppyEnableBehavior(PlatformEnableBehavior):
    def __init__(self, parent):
        super().__init__(parent, platforms=FLOPPY_PLATFORMS)


class CDEnableBehavior(PlatformEnableBehavior):
    def __init__(self, parent):
        super().__init__(parent, platforms=CD_PLATFORMS)


class PlatformShowBehavior(PlatformBehavior):
    def __init__(self, parent, platforms):
        super().__init__(parent, platforms)

    def perform(self, match):
        # FIXME: Temporarily disabled due to having to handle layout
        # updates when widgets are shown/hidden
        widget = self._parent()
        if match:
            widget.show()
            # Workaround for (possibly) a bug where the widget is still
            # disabled if it was enabled while hidden?
            # if widget.is_enabled():
            #     widget.disable()
            #     widget.enable()
        else:
            widget.hide()
        # FIXME: Disabling/enabling instead, for now
        # widget = self._parent()
        # if match:
        #     widget.enable()
        # else:
        #     widget.disable()
        # pass


class AmigaShowBehavior(PlatformShowBehavior):
    def __init__(self, parent):
        super().__init__(parent, platforms=AMIGA_PLATFORMS)



import fsui
from ..skin import Skin


class ConfigScrollArea(fsui.VerticalScrollArea):
    def __init__(self, parent):
        fsui.VerticalScrollArea.__init__(self, parent)
        Skin.set_background_color(self)



import os
import shutil
import tempfile
import threading
import time
import traceback
import weakref
from fsbc.paths import Paths
from fsbc.util import unused
from fsgs.archive import Archive
from fsgs.BaseContext import BaseContext
from fsgs.Database import Database
from fsgs.filedatabase import FileDatabase
from fsgs.GameDatabase import GameDatabase, IncompleteGameException
from fsgs.LockerDatabase import LockerDatabase
from fsgs.download import Downloader, offline_mode
from fsgs.network import is_http_url
from fsgs.ogd.locker import is_locker_enabled, open_locker_uri
from fsgs.plugins.pluginmanager import PluginManager


class NotFoundError(RuntimeError):
    pass


class File(object):
    def __init__(self, path):
        self.path = path


class FileContext(BaseContext):
    def __init__(self, main_context):
        BaseContext.__init__(self, main_context)

    def find_by_sha1(self, sha1):
        # FIXME: check_sha1 should check with PluginManager directly?
        database = FileDatabase.instance()
        result = database.find_file(sha1=sha1)["path"]
        if not result:
            path = Downloader.get_cache_path(sha1)
            if os.path.exists(path):
                result = path
        #    result = self.context.get_game_database().find_file_by_sha1(sha1)
        # print("find by sha1", sha1, "in file database - result", result)
        if not result and is_locker_enabled() and not offline_mode():
            database = LockerDatabase.instance()
            if database.check_sha1(sha1):
                result = "locker://" + sha1
            # print("find by sha1", sha1, "in locker database - result",
            # result)
        return result

    def check_sha1(self, sha1):
        # FIXME: check_sha1 should check with PluginManager directly?
        database = FileDatabase.instance()
        result = database.check_sha1(sha1)
        if not result and is_locker_enabled():
            database = LockerDatabase.instance()
            result = database.check_sha1(sha1)
            # print("check sha1", sha1, "in locker database - result", result)
        # if not result:
        #     result = self.context.get_game_database().find_file_by_sha1(sha1)
        return result

    def get_license_code_for_url(self, url):
        return self.context.get_game_database().get_license_code_for_url(url)

    # FIXME: better name
    def convert_uri(self, uri, prefer_path=False):
        if uri.startswith("sha1://"):
            return self.open_sha1_uri(uri)
        elif uri.startswith("db://"):
            # old name for sha1://
            return self.open_sha1_uri(uri)
        elif is_http_url(uri):
            return self.open_url(uri)
        elif uri.startswith("locker://"):
            return open_locker_uri(uri)
        else:
            if uri.startswith("$"):
                uri = Paths.expand_path(uri)
            if prefer_path and os.path.exists(uri):
                # return helper object so isinstance does not match with str
                return File(uri)
            return Archive(uri).open(uri)

    def open(self, uri, prefer_path=False):
        print("[FILES] Open", uri)
        while isinstance(uri, str):
            uri = self.convert_uri(uri, prefer_path=prefer_path)
        if prefer_path and isinstance(uri, File):
            # is a path
            return uri.path
        elif hasattr(uri, "read"):
            # is a file object
            return uri
        elif uri is None:
            # file was not found
            return None
        raise Exception("unexpected result in fsgs.file.open")

    def open_sha1_uri(self, uri):
        sha1 = uri.split("/")[2]
        assert len(sha1) == 40
        return self.find_by_sha1(sha1)

    def open_url(self, url):
        original_url = url
        hash_part = ""
        parts = url.split("#", 1)
        if len(parts) > 1:
            url = parts[0]
            hash_part = "#" + parts[1]
        if not Downloader.cache_file_from_url(url, download=False):
            license_code = self.get_license_code_for_url(original_url)
            license_status = {"accepted": False, "done": False}

            def show_license_code():
                try:
                    try:
                        license_status["accepted"] = self.show_license_code(
                            license_code
                        )
                    except Exception:
                        traceback.print_exc()
                finally:
                    license_status["done"] = True

            if license_code:
                print("URL", url, "has license code", license_code)
                # FIXME: remove direct dependency on fsui
                import fsui as fsui

                fsui.call_after(show_license_code)
                while not license_status["done"]:
                    time.sleep(0.1)
                if not license_status["accepted"]:
                    # FIXME: custom exception here
                    raise Exception(
                        'Usage terms "{0}" was not '
                        "accepted".format(license_code)
                    )
        path = Downloader.cache_file_from_url(url)
        return path + hash_part

    def show_license_code(self, license_code):
        if license_code == "BTTR":
            license_text = (
                "Files for this game are provided by back2roots.org.\n\n"
                "By using back2roots.org or any of their services you "
                "agree to their Acceptable Usage Policy:\n\n"
                "http://www.back2roots.org/About/Project/Policy/"
            )
        else:
            license_text = license_code
        return self.on_show_license_information(license_text)

    def on_show_license_information(self, license_text):
        unused(license_text)
        print("*** on_show_license_information not implemented ***")
        raise Exception("on_show_license_information not implemented")

    def copy_game_file(self, src, dst):
        try:
            return self._copy_game_file(src, dst)
        except NotFoundError as e:
            if self.context.config.get("download_file"):
                # we should be able to find all missing files after we have
                # downloaded and extracted this archive
                self.download_game_file_archive(
                    self.context.config.get("download_file")
                )
                # now try to re-open the file (should be found in the cache
                return self._copy_game_file(src, dst)
            raise e

    def _copy_game_file(self, src, dst):
        ifs = self.open(src, prefer_path=True)
        if not ifs:
            raise NotFoundError("Could not find file for {0}".format(src))

        if os.path.exists(dst):
            print("removing existing file", dst)
            os.remove(dst)

        if isinstance(ifs, str):
            # we got a direct path
            try:
                os.link(ifs, dst)
                return
            except Exception:
                # couldn't link file, normal on non-unix systems and also
                # if the files are on different file systems
                pass
            shutil.copyfile(ifs, dst)
        else:
            dst_partial = dst + ".partial"
            with open(dst_partial, "wb") as ofs:
                # ifs_sha1 = hashlib.sha1()
                while True:
                    # noinspection PyUnresolvedReferences
                    data = ifs.read()
                    if not data:
                        break
                    # ifs_sha1.update(data)
                    ofs.write(data)
            print("rename file from", dst_partial, "to", dst)
            os.rename(dst_partial, dst)

    def download_game_file_archive(self, url):
        print("\ndownload_game_file_archive", url)
        archive_path = Downloader.cache_file_from_url(url)
        archive = Archive(archive_path)
        archive_files = archive.list_files()
        print(archive_files)
        for name in archive_files:
            print(name)
            ifs = archive.open(name)
            data = ifs.read()
            Downloader.cache_data(data)
        if len(archive_files) == 0:
            # might not be an archive then
            with open(archive_path, "rb") as f:
                data = f.read()
            Downloader.cache_data(data)
        # the downloaded archive is no longer needed, now that we have
        # extracted all the files
        os.remove(archive_path)
        print("\n")


class FSGameSystemContext(object):
    def __init__(self):
        self._amiga = None
        self._config = None
        self._settings = None
        self._signal = None
        self._netplay = None
        self._game = None
        self._plugins = None
        # self._variant = None
        self.file = FileContext(self)
        self.thread_local = threading.local()

    @property
    def amiga(self):
        if self._amiga is None:
            from .amiga.amigacontext import AmigaContext

            self._amiga = AmigaContext(self)
        return self._amiga

    @property
    def config(self):
        if self._config is None:
            from .Config import Config

            self._config = Config(self)
        return self._config

    @property
    def settings(self):
        if self._settings is None:
            from fsbc.settings import Settings

            self._settings = Settings.instance()
        return self._settings

    @property
    def game(self):
        if self._game is None:
            self._game = GameContext(self)
        return self._game

    @property
    def plugins(self):
        if self._plugins is None:
            self._plugins = PluginsContext(self)
        return self._plugins

    # @property
    # def variant(self):
    #     if self._variant is None:
    #         self._variant = VariantContext(self)
    #     return self._variant

    @property
    def signal(self):
        if self._signal is None:
            from .SignalContext import SignalContext

            self._signal = SignalContext(self)
            # self._signal = Signal()
        return self._signal

    @property
    def netplay(self):
        if self._netplay is None:
            from .netplay.NetplayContext import NetplayContext

            self._netplay = NetplayContext(self)
        return self._netplay

    def database(self):
        return Database.instance()

    def get_game_database(self):
        return self.game_database("Amiga")

    def game_database(self, database_name):
        attr_name = "game_database_" + database_name.replace("/", "_")
        if not hasattr(self.thread_local, attr_name):
            # FIXME
            from fsgs.FSGSDirectories import FSGSDirectories

            # FIXME
            # path = os.path.join(
            #     FSGSDirectories.get_cache_dir(), "openretro.org")
            path = os.path.join(
                FSGSDirectories.databases_dir(), database_name + ".sqlite"
            )
            if not os.path.exists(os.path.dirname(path)):
                os.makedirs(os.path.dirname(path))
            # path = os.path.join(path, short_platform_id + ".sqlite")
            database = GameDatabase(path)
            setattr(self.thread_local, attr_name, database)
        return getattr(self.thread_local, attr_name)

    @property
    def cache_dir(self):
        # FIXME: remove dependency
        from fsgs.FSGSDirectories import FSGSDirectories

        return FSGSDirectories.get_cache_dir()

    # def temp_dir(self, suffix):
    #     return TemporaryDirectory(suffix)

    # def temp_file(self, suffix):
    #     return TemporaryFile(suffix)

    def load_game_by_uuid(self, game_uuid):
        if self.load_game_variant(game_uuid):
            return True
        variant_uuid = self.find_preferred_game_variant(game_uuid)
        return self.load_game_variant(variant_uuid)

    # noinspection PyMethodMayBeStatic
    def find_preferred_game_variant(self, game_uuid):
        print("find_preferred_game_variant game_uuid =", game_uuid)
        return self.get_ordered_game_variants(game_uuid)[0]["uuid"]

    # noinspection PyMethodMayBeStatic
    def get_ordered_game_variants(self, game_uuid):
        print("get_ordered_game_variants game_uuid =", game_uuid)
        from .Database import Database

        database = Database.instance()
        variants = database.find_game_variants_new(game_uuid=game_uuid)
        print(variants)
        # ordered_list = []
        # FIXME: Merge code with VariantsBrowser.py
        sortable_items = []
        for i, variant in enumerate(variants):
            game_database = self.game_database(variant["database"])
            variant["like_rating"], variant[
                "work_rating"
            ] = game_database.get_ratings_for_game(variant["uuid"])
            variant[
                "personal_rating"
            ], ignored = database.get_ratings_for_game(variant["uuid"])
            # variant_uuid = variant["uuid"]

            name = variant["name"]
            # FIXME: name replacement needed any more?
            name = name.replace("\n", " (")
            name = name.replace(" \u00b7 ", ", ")
            name += ")"

            if variant["published"] == 0:
                primary_sort = 1
                variant["name"] = "[UNPUBLISHED] " + variant["name"]
            else:
                primary_sort = 0
            sort_key = (
                primary_sort,
                1000000 - variant["like_rating"],
                1000000 - variant["work_rating"],
                name,
            )
            sortable_items.append((sort_key, i, variant))
        ordered_list = [x[2] for x in sorted(sortable_items)]
        print("ordered variant list:")
        for variant in ordered_list:
            print("-", variant["name"])
        # item.configurations = [x[1] for x in ordered_list]
        select_index = None
        if select_index is None:
            # default index selection
            for i, variant in enumerate(ordered_list):
                if variant["personal_rating"] == 5:
                    select_index = i
                    break
            else:
                for i, variant in enumerate(ordered_list):
                    if variant["have"] >= 3:
                        select_index = i
                        break
                else:
                    for i, variant in enumerate(ordered_list):
                        if variant["have"] >= 1:
                            select_index = i
                            break
                    else:
                        if len(ordered_list) > 0:
                            select_index = 0
        # return ordered_list[select_index]["uuid"]
        if select_index and select_index > 0:
            ordered_list.insert(0, ordered_list.pop(select_index))
        return ordered_list

    def load_game_variant(self, variant_uuid):
        # game_database = fsgs.get_game_database()
        # values = game_database.get_game_values_for_uuid(variant_uuid)

        from .Database import Database

        database = Database.instance()
        try:
            database_name = database.find_game_database_for_game_variant(
                variant_uuid
            )
        except LookupError:
            return False

        try:
            values = self.game.set_from_variant_uuid(
                database_name, variant_uuid
            )
        except KeyError:
            # It is possible that the variant is found without game entry,
            # which raises a KeyError.
            return False
        if not values:
            return False

        # print("")
        # for key in sorted(values.keys()):
        #     print(" * {0} = {1}".format(key, values[key]))
        # print("")

        from fsgs.platform import PlatformHandler

        platform_handler = PlatformHandler.create(self.game.platform.id)
        loader = platform_handler.get_loader(self)
        self.config.load(loader.load_values(values))
        return True

    def run_game(self):
        from fsgs.platform import PlatformHandler

        platform_handler = PlatformHandler.create(self.game.platform.id)
        runner = platform_handler.get_runner(self)

        from fsgs.input.enumeratehelper import EnumerateHelper

        device_helper = EnumerateHelper()
        device_helper.default_port_selection(runner.ports, runner.options)

        runner.prepare()
        process = runner.run()
        process.wait()
        runner.finish()


# class TemporaryDirectory(object):
#
#     def __init__(self, suffix):
#         self.path = tempfile.mkdtemp(suffix="-fsgs-" + suffix)
#
#     def __del__(self):
#         self.delete()
#
#     def delete(self):
#         if os.environ.get("FSGS_CLEANUP", "") == "0":
#             print("NOTICE: keeping temp files around...")
#             return
#         if self.path:
#             shutil.rmtree(self.path)
#             self.path = ""


# class TemporaryFile(object):
#
#     def __init__(self, suffix):
#         # self.path = tempfile.mkstemp(suffix=suffix)
#         self.dir_path = tempfile.mkdtemp(suffix="-fsgs-" + suffix)
#         self.path = os.path.join(self.dir_path, suffix)
#
#     def __del__(self):
#         self.delete()
#
#     def delete(self):
#         if os.environ.get("FSGS_CLEANUP", "") == "0":
#             print("NOTICE: keeping temp files around...")
#             return
#         if self.path:
#             os.unlink(self.path)
#             self.path = ""
#         if self.dir_path:
#             shutil.rmtree(self.dir_path)
#             self.dir_path = ""


class GameContext(object):
    def __init__(self, context):
        self._context = weakref.ref(context)
        # FIXME: REMOVE? use config[game_name] instead?
        self.name = ""
        # FIXME: REMOVE? use config[game_uuid] instead?
        self.uuid = ""
        self.variant = VariantContext()
        self.platform = GamePlatform()

    @property
    def fsgs(self):
        return self._context()

    def get_game_values_for_id(self, game_database, id):
        doc = game_database.get_game_values(id)
        return self._fix_with_game_values_from_parent_database(doc)

    def get_game_values_for_uuid(self, game_database, uuid):
        doc = game_database.get_game_values_for_uuid(uuid)
        return self._fix_with_game_values_from_parent_database(doc)

    def _fix_with_game_values_from_parent_database(self, doc):
        parent_database_name = doc.get("parent_database", "")
        if parent_database_name:
            parent_database = self.fsgs.game_database(parent_database_name)
            try:
                parent_doc = parent_database.get_game_values(
                    game_uuid=doc.get("parent_uuid", "")
                )
            except LookupError:
                raise IncompleteGameException(
                    "Could not find parent {0} of game {1}".format(
                        doc.get("parent_uuid", ""), doc.get("game_uuid")
                    )
                )
            parent_doc.update(doc)
            doc = parent_doc
        return doc

    def set_from_variant_uuid(self, database_name, variant_uuid):
        print("set_from_variant_uuid", database_name, variant_uuid)
        game_database = self.fsgs.game_database(database_name)
        doc = self.get_game_values_for_uuid(game_database, variant_uuid)

        if not doc.get("_type", "") == "2":
            return {}
        print("")
        for key in sorted(doc.keys()):
            print(" * {0} = {1}".format(key, doc[key]))
        print("")

        platform_id = doc["platform"]
        self.platform.id = platform_id

        self.uuid = doc["game_uuid"]
        self.name = doc["game_name"]
        self.variant.uuid = variant_uuid
        self.variant.name = doc["variant_name"]
        return doc


class PluginsContext(object):
    def __init__(self, context):
        self._context = weakref.ref(context)

    def find_resource(self, name):
        return PluginManager.instance().find_resource(name)

    def find_executable(self, name):
        return PluginManager.instance().find_executable(name)


class GamePlatform(object):
    def __init__(self):
        self._id = ""

    def get_id(self):
        return self._id

    def set_id(self, id):
        self._id = id.lower()

    id = property(get_id, set_id)

    @property
    def name(self):
        from .platform import PlatformHandler

        return PlatformHandler.get_platform_name(self._id)
        # if self._id == "atari-7800":
        #     return "Atari 7800"
        # if self._id == "amiga":
        #     return "Amiga"
        # if self._id == "cdtv":
        #     return "CDTV"
        # if self._id == "cd32":
        #     return "CD32"
        # raise Exception("Unrecognized platform ({0})".format(self._id))


class VariantContext(object):
    def __init__(self):
        self.name = ""
        self.uuid = ""



import os
import io
import xml.etree.cElementTree as ElementTree


class DatFile(object):
    def __init__(self, file=None):
        self.games = []
        self.description = ""

        self.extensions = []
        if file is not None:
            self.load(file)

    def reset(self):
        self.games = []
        self.description = ""

    def load(self, file):
        if hasattr(file, "read"):
            self._load_file(file)
        else:
            self._load_path(file)

    def _load_path(self, path):
        with io.open(path, "r", encoding="UTF-8") as f:
            self._load_file(f)

    def _load_file(self, f):
        self.load_data(f.read())

    def load_data(self, data):
        if data.startswith("<") or data.startswith("\uFEFF"):
            # assume XML
            self._load_xml(data)
        else:
            self._load_dat(data)

    def _add_game_rom(self, game, rom_node):
        # print(rom_node.attrib)
        rom = {
            "name": rom_node.attrib["name"],
            "size": int(rom_node.attrib["size"]),
        }
        if "crc" in rom_node.attrib:
            rom["crc32"] = rom_node.attrib["crc"].lower()
        if "md5" in rom_node.attrib:
            rom["md5"] = rom_node.attrib["md5"].lower()
        if "sha1" in rom_node.attrib:
            rom["sha1"] = rom_node.attrib["sha1"].lower()
        game["files"].append(rom)

    def _add_game(self, node):
        game = {"name": node.attrib["name"], "files": []}
        for rom_node in node.findall("rom"):
            self._add_game_rom(game, rom_node)
        self.games.append(game)

    def _load_mame(self, root):
        for machine_node in root.findall("machine"):
            self._add_game(machine_node)

    def _load_softwarelists(self, root):
        for softwarelist_node in root.findall("softwarelist"):
            # FIXME: software list name
            for software_node in softwarelist_node.findall("software"):
                # FIXME: software list name + software name?
                name = software_node.attrib["name"]
                game = {"name": name, "files": []}
                for rom_node in software_node.iter("rom"):
                    if rom_node.attrib.get("name", ""):
                        self._add_game_rom(game, rom_node)
                self.games.append(game)

    def _load_xml(self, data):
        self.reset()

        root = ElementTree.fromstring(data)
        if root.tag == "mame":
            return self._load_mame(root)
        if root.tag == "softwarelists":
            return self._load_softwarelists(root)
        header_node = root.find("header")
        description_node = header_node.find("description")
        if description_node is not None:
            self.description = description_node.text.strip()
        for game_node in root.findall("game"):
            self._add_game(game_node)

    def _load_dat(self, data):
        self.reset()

        # this is a very quick-and-dirty/hackish parser...
        game = None
        for line in data.split("\n"):
            line = line.strip()
            if line.startswith("game ("):
                game = {"name": "", "files": []}
                self.games.append(game)
            if line.startswith("name "):
                if game is not None:
                    if '"' in line:
                        parts = line.split('"')
                        assert len(parts) == 3
                        game["name"] = parts[1]
                    else:
                        game["name"] = line[5:]
            elif line.startswith("description "):
                if game is None:
                    parts = line.split('"')
                    assert len(parts) == 3
                    self.description = parts[1]
            elif line.startswith("rom ("):
                rom = {}
                parts = line.split('"')
                if len(parts) == 3:
                    rom["name"] = parts[1]
                else:
                    parts = line.split(" ")
                    rom["name"] = parts[3]
                n, ext = os.path.splitext(rom["name"])
                if self.extensions and ext.lower() not in self.extensions:
                    continue
                parts = line.split(" size ")
                rom["size"] = int(parts[1].strip().split(" ")[0].strip())
                parts = line.split(" crc ")
                rom["crc32"] = parts[1].strip().split(" ")[0].strip().lower()
                parts = line.split(" md5 ")
                rom["md5"] = parts[1].strip().split(" ")[0].strip().lower()
                parts = line.split(" sha1 ")
                rom["sha1"] = parts[1].strip().split(" ")[0].strip().lower()
                game["files"].append(rom)



# FSGS - Common functionality for FS Game System.
# Copyright (C) 2013-2019  Frode Solheim <frode@solheim.dev>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
from fsgs.knownfiles import KnownFile
from fsgs.platforms import PLATFORM_ZXS

ZXS_48_ROM = KnownFile(
    "5ea7c2b824672e914525d1d5c419d71b84a426a2", PLATFORM_ZXS, "48.rom"
)
ZXS_128_0_ROM = KnownFile(
    "4f4b11ec22326280bdb96e3baf9db4b4cb1d02c5", PLATFORM_ZXS, "128-0.rom"
)
ZXS_128_1_ROM = KnownFile(
    "80080644289ed93d71a1103992a154cc9802b2fa", PLATFORM_ZXS, "128-1.rom"
)
ZXS_PLUS3_0_ROM = KnownFile(
    "e319ed08b4d53a5e421a75ea00ea02039ba6555b", PLATFORM_ZXS, "plus3-0.rom"
)
ZXS_PLUS3_1_ROM = KnownFile(
    "c9969fc36095a59787554026a9adc3b87678c794", PLATFORM_ZXS, "plus3-1.rom"
)
ZXS_PLUS3_2_ROM = KnownFile(
    "22e50c6ba4157a3f6a821bd9937cd26e292775c6", PLATFORM_ZXS, "plus3-2.rom"
)
ZXS_PLUS3_3_ROM = KnownFile(
    "65f031caa8148a5493afe42c41f4929deab26b4e", PLATFORM_ZXS, "plus3-3.rom"
)



import os
import shutil

from fsgs.drivers.mamedriver import MameDriver
from fsgs.saves import SaveHandler


class MessDriver(MameDriver):
    def __init__(self, fsgs):
        super().__init__(fsgs)
        self.save_handler = MessSaveHandler(self.fsgc, options=self.options)

    def prepare(self):
        super().prepare()
        self.save_handler.prepare()
        self.create_mame_layout()

    def finish(self):
        super().finish()
        self.save_handler.finish()

    def is_pal(self):
        # return self.config.get("ntsc_mode") != "1"
        # return False
        refresh_rate = self.get_game_refresh_rate()
        if refresh_rate:
            return int(round(refresh_rate)) == 50

    def mame_emulator_name(self):
        return "mess"

    def mame_init_input(self):
        self.mess_init_input()

    def mess_init_input(self):
        pass

    def mame_romset(self):
        result = {}
        romset_name, files = self.mess_romset()
        for sha1, name in files.items():
            result[name] = sha1
        return romset_name, result

    def mess_romset(self):
        return self.get_mess_romset()

    def get_mess_romset(self):
        return "", {}

    # def mess_get_firmware_name(self):
    #     return self.context.game.platform + ' Firmware'

    def mame_prepare(self):
        pass
        # firmware_name = self.mess_get_firmware_name()
        # if firmware_name:
        #     self.mess_firmware_dir = self.prepare_firmware(firmware_name)
        # else:
        #     self.mess_firmware_dir = None

    def mame_configure(self):
        # self.add_arg(self.mess_romset()[0])
        # self.configure_mame()
        # with open(self.mednafen_cfg_path(), 'wb') as f:
        #     self._configure_emulator(f)
        # bios_dir = self.get_bios_dir()
        # if bios_dir:
        #     self.args.extend(['-rompath', bios_dir])
        # self.args.extend(['-joystick_deadzone', '0.15'])

        self.mess_configure()
        # self.mess_configure_media()

        if not self.mess_full_keyboard():
            # Start mess with UI keyboard keys enabled by default.
            # Full keyboard can be activated with INS / Scroll-Lock
            self.emulator.args.extend(["-ui_active"])

    def mess_full_keyboard(self):
        return False

    def mess_configure(self):
        """override in subclasses to provide custom configuration"""
        pass

    def mame_input_mapping(self, port):
        return self.mess_input_mapping(port)

    def mess_input_mapping(self, port):
        return self.get_mess_input_mapping(port)

    def get_mess_input_mapping(self, port):
        return None

    def mame_get_bios_dir(self):
        print("[DRIVER] MESS Firmware Dir:", self.mess_firmware_dir)
        return self.mess_firmware_dir

    def mess_offset_and_scale(self):
        return None

    def mame_offset_and_scale(self):
        result = self.mess_offset_and_scale()
        if result is not None:
            return result
        return super().mame_offset_and_scale()

    #    def mess_configure_media(self):
    #        pass

    def mess_configure_cartridge(self, slot="cart"):
        self.add_arg("-{0}".format(slot), self.get_game_file())

    def mess_configure_floppies(self, slots):
        # FIXME: this is a quick hack only..
        self.add_arg("-{0}".format(slots[0]), self.get_game_file())

    def inject_fake_input_string_list(self, delay, inject_string):
        self.set_env("FSGS_FAKE_INPUT_DELAY", str(delay))
        self.set_env("FSGS_FAKE_INPUT", "".join(inject_string))

    def inject_fake_input_string(self, delay, inject_string):
        def inject(inject_key_code):
            s.append("1{0:03d}".format(inject_key_code))
            s.append("0{0:03d}".format(inject_key_code))
            # we add some dummy events to slow down the keyboard entry

            # s.append("1000")
            s.append("0000")

        s = []
        for c in inject_string:
            if c == '"':
                # s.append("1304" + "1050" + "0050" + "0304")
                s.append("1225" + "1031" + "0031" + "0225")
            elif c in " ":
                # inject(32)
                inject(44)
            elif c in "\n":
                # inject(13)
                inject(40)
            elif c == "-":
                # inject(45)
                inject(45)
            elif c in "0123456789":
                # code = 48 + "0123456789".index(c)
                code = 30 + "1234567890".index(c)
                inject(code)
            elif c in "abcdefghijklmnopqrstuvwxyz":
                # code = 97 + "abcdefghijklmnopqrstuvwxyz".index(c)
                code = 4 + "abcdefghijklmnopqrstuvwxyz".index(c)
                inject(code)
            else:
                raise Exception(
                    "inject_fake_input_string cannot "
                    "handle '{0}' yet".format(c)
                )
        self.inject_fake_input_string_list(delay, s)


class MessSaveHandler(SaveHandler):
    def __init__(self, fsgc, options):
        super().__init__(fsgc, options, emulator="MAME")
        self._mame_driver = ""

    def prepare(self):
        if self._emulator_specific:
            if not self._mame_driver:
                raise Exception("MAME driver not specified")
            self.copy_to_mame_dir()
        super().prepare()

    def finish(self):
        if self._emulator_specific:
            self.move_from_mame_dir()
        super().finish()

    def set_mame_driver(self, mame_driver):
        self._mame_driver = mame_driver

    def mame_save_dirs(self):
        save_dir = self.emulator_save_dir()
        mame_save_dir = os.path.join(save_dir, self._mame_driver)
        return save_dir, mame_save_dir

    def copy_to_mame_dir(self):
        save_dir, mame_save_dir = self.mame_save_dirs()
        if not os.path.exists(save_dir):
            return
        for full_name in os.listdir(save_dir):
            src = os.path.join(save_dir, full_name)
            if not os.path.isfile(src):
                continue
            name, ext = os.path.splitext(full_name)
            if ext not in [".srm"]:
                continue
            if not os.path.exists(mame_save_dir):
                os.makedirs(mame_save_dir)
            dst = os.path.join(mame_save_dir, name + ".nv")
            print("MessSaveHandler:", src, "->", dst)
            shutil.copy(src, dst)

    def move_from_mame_dir(self):
        save_dir, mame_save_dir = self.mame_save_dirs()
        if not os.path.exists(mame_save_dir):
            return
        for full_name in os.listdir(mame_save_dir):
            src = os.path.join(mame_save_dir, full_name)
            if not os.path.isfile(src):
                continue
            name, ext = os.path.splitext(full_name)
            if ext not in [".nv"]:
                print("MessSaveHandler: Removing", src)
                os.remove(src)
                continue
            dst = os.path.join(save_dir, name + ".srm")
            print("MessSaveHandler:", dst, "<-", src)
            shutil.copy(src, dst)
            os.remove(src)
        try:
            # FIXME: Maybe try to force removing this directory, at least
            # if we created it, and self._mame_driver is specified
            os.rmdir(mame_save_dir)
        except Exception:
            pass



import sys
from fsgs.plugins.pluginmanager import PluginManager

"""
RetroArch launcher script used for testing.
"""


# FIXME: Use from RetroArchDriver
def find_libretro_core(name):
    return "/usr/lib/x86_64-linux-gnu/libretro/{}.so".format(name)


def app_main():
    executable = PluginManager.instance().find_executable("retroarch")

    args = sys.argv[1:]
    libretro_core = find_libretro_core("nestopia_libretro")
    args.extend(["-L", libretro_core])

    process = executable.popen(args)
    process.wait()



import os

from fsbc.paths import Paths
from fsgs.archive import Archive
from fsgs.checksumtool import ChecksumTool
from fsgs.FSGSDirectories import FSGSDirectories
from fsgs.amiga.amiga import Amiga
from fsgs.context import fsgs
from .i18n import gettext
from .launcher_config import LauncherConfig
from .ui.LauncherFilePicker import LauncherFilePicker


class FloppyManager(object):
    @classmethod
    def clear_all(cls):
        for i in range(4):
            cls.eject(i)
        cls.clear_floppy_list()

    @classmethod
    def eject(cls, drive):
        # values = []
        # values.append(("floppy_drive_{0}".format(drive), ""))
        # values.append(("x_floppy_drive_{0}_sha1".format(drive), ""))
        # Config.set_multiple(values)
        fsgs.amiga.eject_floppy(drive)

    @classmethod
    def clear_floppy_list(cls):
        values = []
        for i in range(Amiga.MAX_FLOPPY_IMAGES):
            values.append(("floppy_image_{0}".format(i), ""))
            values.append(("x_floppy_image_{0}_sha1".format(i), ""))
        LauncherConfig.set_multiple(values)

    @classmethod
    def multi_select(cls, parent=None):
        default_dir = FSGSDirectories.get_floppies_dir()
        dialog = LauncherFilePicker(
            parent,
            gettext("Select Multiple Floppies"),
            "floppy",
            multiple=True,
        )
        if not dialog.show_modal():
            return
        original_paths = dialog.get_paths()
        original_paths.sort()
        paths = []
        for path in original_paths:
            path = Paths.get_real_case(path)
            embedded_files = []
            if path.endswith(".zip"):
                archive = Archive(path)
                files = archive.list_files()
                for file in files:
                    name, ext = os.path.splitext(file)
                    # FIXME: get list of floppy extensions from a central
                    # place
                    if ext in [".adf", ".ipf"]:
                        embedded_files.append(file)
            if len(embedded_files) > 0:
                embedded_files.sort()
                print("found embedded floppy images:")
                print(embedded_files)
                for file in embedded_files:
                    paths.append(file)
            else:
                paths.append(path)

        checksum_tool = ChecksumTool(parent)
        for i, path in enumerate(paths):
            sha1 = checksum_tool.checksum(path)
            path = Paths.contract_path(
                path, default_dir, force_real_case=False
            )

            if i < 4:
                LauncherConfig.set_multiple(
                    [
                        ("floppy_drive_{0}".format(i), path),
                        ("x_floppy_drive_{0}_sha1".format(i), sha1),
                    ]
                )
            LauncherConfig.set_multiple(
                [
                    ("floppy_image_{0}".format(i), path),
                    ("x_floppy_image_{0}_sha1".format(i), sha1),
                ]
            )

        # blank the rest of the drives
        for i in range(len(paths), 4):
            LauncherConfig.set_multiple(
                [
                    ("floppy_drive_{0}".format(i), ""),
                    ("x_floppy_drive_{0}_sha1".format(i), ""),
                ]
            )
        # blank the rest of the image list
        for i in range(len(paths), 20):
            LauncherConfig.set_multiple(
                [
                    ("floppy_image_{0}".format(i), ""),
                    ("x_floppy_image_{0}_sha1".format(i), ""),
                ]
            )
            # dialog.destroy()



import sys
import time
import weakref

import fsui
from arcade.callbacks import Callbacks
from arcade.glui.imageloader import ImageLoader
from arcade.glui.input import InputHandler
from arcade.ui.event import Event
from arcade.ui.gl_widget import GLWidget
from fsbc.settings import Settings
from fsbc.system import macosx
from fsui.qt import init_qt, Qt, QWidget, QKeyEvent

CURSOR_SHOW_DURATION = 5.0


def check_argument(name, options=None):
    name = name.replace("_", "-")
    if options is not None:
        for option in options:
            if "--" + name + "=" + option in sys.argv:
                return option
    if "--" + name in sys.argv:
        return "1"
    if "--" + name + "=1" in sys.argv:
        return "1"
    if "--no-" + name in sys.argv:
        return "0"
    if "--" + name + "=0" in sys.argv:
        return "0"
    return ""


def fullscreen():
    # If we have explicitly used --window as arguments, do
    # not enable fullscreen regardless of settings.
    if check_argument("window", ["maximize"]) in ["1", "maximize"]:
        return False
    # if check_argument("maximize") == "1":
    #     return False
    value = check_argument("fullscreen")
    if not value:
        value = Settings.instance().get("arcade_fullscreen")
    return value != "0"


def maximized():
    # if check_argument("fullscreen") == "1":
    #     return False
    # if check_argument("window") == "1":
    #     return False
    # value = check_argument("window") == "maximize"
    value = check_argument("window", ["maximize"]) == "maximize"
    if value:
        return True
    else:
        value = Settings.instance().get("arcade_maximized")
        return value == "1"


def monitor():
    value = check_argument("monitor")
    if not value:
        value = Settings.instance().get("monitor")
    if not value:
        value = "middle-left"
    return value


def decorations():
    return not maximized()


def screen_geometry():
    q_app = init_qt()
    desktop = q_app.desktop()
    screens = []
    for i in range(desktop.screenCount()):
        geom = desktop.screenGeometry(i)
        screens.append([geom.x(), i, geom])
    screens.sort()
    if monitor() == "left":
        mon = 0
    elif monitor() == "middle-right":
        mon = 2
    elif monitor() == "right":
        mon = 3
    else:  # middle-left
        mon = 1
    display = round(mon / 3 * (len(screens) - 1))
    geom = screens[display][2]
    return geom.x(), geom.y(), geom.width(), geom.height()

    # main_window.setGeometry(geometry)


class ArcadeWindow(fsui.Window):
    @classmethod
    def open(cls, parent=None):
        return fsui.open_window_instance(cls, parent)

    def __init__(self, parent=None):
        # if app.name == "fs-uae-arcade":
        title = "FS-UAE Arcade"
        # else:
        #    title = "FS Game Center"

        border = True
        if maximized():
            border = False

        super().__init__(
            parent,
            title,
            separator=False,
            border=border,
            menu=True,
            color=(0x00, 0x00, 0x00),
        )
        self.set_background_color(fsui.Color(0x00, 0x00, 0x00))
        self.layout = fsui.HorizontalLayout()
        self.quit_flag = False
        callbacks = Callbacks()
        callbacks.set_window(self)

        interval = 16
        self.qt_window = QtWindow(callbacks, interval, window=self)
        self.adapter = fsui.Adapter(self, self.qt_window)
        self.adapter.set_min_size((960, 540))
        self.layout.add(self.adapter, expand=True, fill=True)

        # self.set_size((960, 540))
        # qt_window.setFocus()
        self.adapter.focus()
        self.shown.connect(self.on_show)
        self.closed.connect(self.on_close)

    def on_show(self):
        self.qt_window.create_gl_window()

    def show_auto(self):
        if fullscreen():
            geometry = screen_geometry()
            self.set_fullscreen(True, geometry)
            Settings.instance().set("__cursor_x", geometry[2])
            Settings.instance().set("__cursor_y", geometry[3])
        elif maximized():
            x, y, w, h = screen_geometry()
            self.set_maximized(True, (x, y, 960, 540))
        else:
            super().show()

    def quit(self):
        self.quit_flag = True
        self.qt_window.quit_flag = True
        # self.window().close()

    def on_close(self):
        # self.qt_window.killTimer(self.qt_window.timer_id)

        # FIXME: In order to clean up resources in the GL thread, we want
        # want to intercept the close event, and close by telling glui to
        # quit, so proper cleanup can be performed there.
        # It may not be a problem though, if the Qt OpenGL context simply
        # releases all resources.
        pass

        ImageLoader.get().stop()


# noinspection PyPep8Naming
class QtWindow(QWidget):
    def __init__(self, callbacks, interval, window):
        super().__init__()
        set_black_background(self)
        self.gl_widget = None
        self.timer_id = None
        self.callbacks = callbacks
        self.interval = interval
        self.quit_flag = False
        self.first_time = None
        # self.setCursor(Qt.BlankCursor)
        self._window = weakref.ref(window)
        self.set_blank_cursor()
        self.show_cursor_until = None
        self.setMouseTracking(True)
        # self.window_created_at = time.time()
        self.first_motion_event = True

    def set_blank_cursor(self, blank=True):
        if blank:
            cursor = Qt.BlankCursor
        else:
            cursor = Qt.ArrowCursor
        self.setCursor(cursor)
        if self.gl_widget is not None:
            self.gl_widget.setCursor(cursor)
        if blank:
            # Fool app to think mouse has moved to neutral position,
            # in order to "de-focus" focused item.
            InputHandler.add_event(
                Event.create_fake_mouse_event(
                    "mouse-motion", 960, 540, (self.width(), self.height())
                )
            )

    def window(self) -> ArcadeWindow:
        # return self.parent().window()
        return self._window()

    def create_gl_window(self):
        self.timer_id = self.startTimer(self.interval)

    def create_gl_window_2(self):
        if self.gl_widget is not None:
            return True
        # Delaying creating of the GLWidget solves some initial sizing
        # issues when maximizing / setting fullscreen on Linux at least.
        # EDIT: The problem may no longer exist, but it is fine to delay
        # anyway so the black screen has time to show before the main thread
        # will block a short while (while resources are loaded).
        if time.time() - self.first_time > 0.5:
            self.gl_widget = GLWidget(self, self.callbacks)
            self.gl_widget.setMouseTracking(True)
            # if "--show-cursor" not in sys.argv:
            # self.gl_widget.setCursor(Qt.BlankCursor)
            self.set_blank_cursor()
            self.gl_widget.setGeometry(
                0, 0, self.size().width(), self.size().height()
            )
            self.gl_widget.show()
        return False

    def restore_window_if_necessary(self):
        pass

    def resizeEvent(self, event):
        size = event.size()
        print("QtWindow.resizeEvent size =", (size.width(), size.height()))
        # if (size.width(), size.height()) == (100, 30):
        # traceback.print_stack()
        if self.gl_widget is not None:
            self.gl_widget.setGeometry(0, 0, size.width(), size.height())

    def timerEvent(self, event):
        if self.first_time is None:
            self.first_time = time.time()
        if not self.create_gl_window_2():
            return
        self.callbacks.timer()
        if self.quit_flag:
            self.killTimer(self.timer_id)
            self.window().close()
            return
        if self.callbacks.active():
            self.gl_widget.updateGL()
        if self.show_cursor_until is not None:
            if self.show_cursor_until < time.time():
                print("hide cursor again")
                self.set_blank_cursor()
                self.show_cursor_until = None

    def ensure_cursor_visible(self):
        if self.show_cursor_until is None:
            print("show cursor until", time.time() + CURSOR_SHOW_DURATION)
            self.set_blank_cursor(False)
        self.show_cursor_until = time.time() + CURSOR_SHOW_DURATION

    def mouseMoveEvent(self, event):
        if self.first_motion_event:
            # Ignore initial motion event, so the cursor is not visible
            # at startup.
            self.first_motion_event = False
            return
        InputHandler.add_event(
            Event.create_mouse_event(event, (self.width(), self.height()))
        )
        self.ensure_cursor_visible()

    def mousePressEvent(self, event):
        if self.show_cursor_until is None:
            # Ignore clicks when cursor is hidden
            return
        InputHandler.add_event(
            Event.create_mouse_event(event, (self.width(), self.height()))
        )
        self.ensure_cursor_visible()

    def mouseReleaseEvent(self, event):
        if self.show_cursor_until is None:
            # Ignore clicks when cursor is hidden
            return
        InputHandler.add_event(
            Event.create_mouse_event(event, (self.width(), self.height()))
        )
        self.ensure_cursor_visible()

    def keyPressEvent(self, event):
        def modifier():
            if macosx:
                # This should correspond to the Cmd key(s) on OS X
                return int(event.modifiers()) & Qt.ControlModifier
            else:
                return int(event.modifiers()) & Qt.AltModifier

        assert isinstance(event, QKeyEvent)
        # print(event.isAutoRepeat(), event.type())
        if event.isAutoRepeat():
            return
        if modifier():
            if event.key() == Qt.Key_Return:
                self.window().set_fullscreen(not self.window().is_fullscreen())
                return
            if event.key() == Qt.Key_Q:
                self.window().close()
                return

        InputHandler.add_event(Event.create_key_event(event))
        text = event.text()
        if text and text in TEXT_WHITE_LIST:
            # We don't want special characters such as return, backspace
            # and escape (etc) to be sent as text events. For now, we use
            # a simple white list.
            InputHandler.add_event({"type": "text", "text": event.text()})

    def keyReleaseEvent(self, event):
        assert isinstance(event, QKeyEvent)
        # print(QKeyEvent, event.isAutoRepeat(), event.type())
        if event.isAutoRepeat():
            return
        InputHandler.add_event(Event.create_key_event(event))


TEXT_WHITE_LIST = (
    "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.,- "
)


def set_black_background(widget):
    palette = widget.palette()
    # FIXME
    palette.setColor(widget.backgroundRole(), Qt.blue)
    widget.setPalette(palette)
    widget.setAutoFillBackground(True)
    widget.setStyleSheet("background-color: black;")



import fsui
from launcher.i18n import gettext
from launcher.launcher_settings import LauncherSettings
from launcher.settings.settings_page import SettingsPage


class VideoSyncSettingsPage(SettingsPage):
    def __init__(self, parent):
        super().__init__(parent)
        icon = fsui.Icon("video-settings", "pkg:workspace")
        gettext("Video Synchronization Settings")
        title = gettext("Synchronization")
        subtitle = gettext(
            "Synchronize FS-UAE with your display for " "smooth video"
        )
        self.add_header(icon, title, subtitle)

        label = fsui.MultiLineLabel(
            self,
            gettext(
                "Enabling the following option will synchronize the emulation "
                "to the display when the emulation refresh rate matches the"
                "screen refresh rate."
            ),
            640,
        )
        self.layout.add(label, fill=True, margin_top=0)
        self.video_sync_group = self.add_option("video_sync")

        self.low_latency_group = self.add_option("low_latency_vsync")

        # label = fsui.MultiLineLabel(self, gettext(
        #     "Enabling the following option will prevent tearing from "
        #     "occurring, but will also use more CPU. Input latency "
        #     "may become slightly higher."), 640)
        # self.layout.add(label, fill=True, margin_top=0)
        # self.vblank_checkbox = fsui.HeadingCheckBox(self, gettext(
        #     "Synchronize buffer swaps with display (prevents tearing)"))
        # self.layout.add(self.vblank_checkbox, margin_top=20)

        self.sync_method_label = fsui.MultiLineLabel(
            self,
            gettext(
                "Depending on your OS and OpenGL drivers, synchronizing "
                "can use needlessly much CPU (esp. applies to "
                "Linux). You can experiment with different sync methods "
                "to improve performance."
            ),
            640,
        )
        self.layout.add(self.sync_method_label, fill=True, margin_top=20)
        self.sync_method_group = self.add_option("video_sync_method")

        # self.smooth_label = fsui.MultiLineLabel(self, gettext(
        #     "In order to get really smooth Amiga graphics, you need to "
        #     "enable the following option, and also make sure your display "
        #     "is running at 50Hz (for PAL) or 60Hz (for NTSC)."), 640)
        # self.layout.add(self.smooth_label, fill=True, margin_top=20)
        # self.full_sync_checkbox = fsui.HeadingCheckBox(self, gettext(
        #     "Also synchronize emulation with display when possible "
        #     "(smooth scrolling)"))
        # self.layout.add(self.full_sync_checkbox, margin_top=20)

        self.layout.add_spacer(0, expand=True)

        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True, margin_top=10)
        hori_layout.add(
            fsui.ImageView(
                self, fsui.Image("launcher:res/16x16/world_link.png")
            )
        )
        label = fsui.URLLabel(
            self,
            gettext("How to achieve perfectly smooth scrolling"),
            "https://fs-uae.net/perfectly-smooth-scrolling",
        )
        hori_layout.add(label, margin_left=6)

        text = gettext(
            "Synchronizing with the display can in some cases cause "
            "increased stuttering and low frame rates (esp. in some Linux "
            "desktop environments with compositing enabled)."
        )
        link = (
            " <a href='https://fs-uae.net/video-synchronization-issues'>"
            "{0}</a>.".format(gettext("Read more"))
        )

        label = fsui.MultiLineLabel(self, text + link, min_width=640)
        self.layout.add(label, fill=True, margin_top=20)

        LauncherSettings.add_listener(self)
        for key in ["video_sync"]:
            self.on_setting(key, LauncherSettings.get(key))

        # self.vblank_checkbox.changed.connect(self.on_vblank_changed)
        # self.full_sync_checkbox.changed.connect(self.on_full_sync_changed)

    def on_destroy(self):
        LauncherSettings.remove_listener(self)

    def on_setting(self, key, _):
        if key == "video_sync":
            # value = value.lower()
            # self.vblank_checkbox.check(value in ["auto", "full", "vblank"])
            # self.full_sync_checkbox.check(value in ["auto", "full"])
            self.update_widgets()

    # def on_vblank_changed(self):
    #     self.on_either_checkbox_changed()
    #
    # def on_full_sync_changed(self):
    #     self.on_either_checkbox_changed()
    #
    # def on_either_checkbox_changed(self):
    #     value = ""
    #     if self.vblank_checkbox.is_checked():
    #         if self.full_sync_checkbox.is_checked():
    #             value = "auto"
    #         else:
    #             value = "vblank"
    #     LauncherSettings.set("video_sync", value)

    def update_widgets(self):
        value = LauncherSettings.get("video_sync")
        vblank = value in ["1", "auto", "full", "vblank"]
        full = value in ["1", "auto", "full"]

        # self.full_sync_checkbox.enable(vblank)
        self.sync_method_group.label.enable(vblank)
        self.sync_method_group.widget.enable(vblank)
        self.sync_method_group.help_button.enable(vblank)
        self.sync_method_label.enable(vblank)
        # self.smooth_label.enable(vblank)
        self.low_latency_group.label.enable(full)
        self.low_latency_group.widget.enable(full)
        self.low_latency_group.help_button.enable(full)



import logging

from pkg_resources import resource_stream

from fsbc.resources import Resources as BaseResources

logger = logging.getLogger("arcade")


def _(msg):
    return str(msg)


def ngettext(n, msg1, msg2):
    return str(msg1) if n == 1 else str(msg2)


class Resources(BaseResources):
    def __init__(self, package):
        BaseResources.__init__(self, package, "res")
        self.req = package

    def resource_pil_image(self, resource):
        # resource_name = encode_path(u'res/' + resource_name)
        # return resource_pil_image(self.req, resource_name)
        stream = self.stream(resource)
        # noinspection PyUnresolvedReferences
        from PIL import Image

        return Image.open(stream)

    def resource_qt_image(self, resource):
        from fsui.qt import QImage

        stream = self.stream(resource)
        im = QImage()
        im.loadFromData(stream.read())
        return im

    def resource_stream(self, resource):
        return self.stream(resource)

    def resource_filename(self, resource):
        return self.path(resource)


def resource_pil_image(package_or_requirement, resource_name):
    # print("resource_pil_image", package_or_requirement, resource_name)
    stream = resource_stream(package_or_requirement, resource_name)
    # noinspection PyUnresolvedReferences
    from PIL import Image

    return Image.open(stream)


# def resource_icon_stream(package_or_requirement, name, size):
#     #print(name, size)
#     resource_name = encode_path(u'res/icons/%dx%d/%s.png' % (size, size,
# name))
#     return resource_stream(package_or_requirement, resource_name)


# def resource_icon_pil(package_or_requirement, name, size):
#     #print("resource_icon_pil", name, size)
#     #path = os.path.join(self.path, "icons",
#     #        "%dx%d" % (size, size), name + ".png")
#     #if os.path.isfile(path):
#     #    return Image.open(path)
#     stream = resource_icon_stream(package_or_requirement, name, size)
#     from PIL import Image
#     return Image.open(stream)

gettext = _

resources = Resources("arcade")



from fsui.qt import QMainWindow, QWidget, Qt
from .gamecenterview import GameCenterView


class FSGSMainWindow(QMainWindow):
    def __init__(self):
        QMainWindow.__init__(self)

        self.game_center_view = GameCenterView()
        self.game_center_view.engine().quit.connect(self.on_quit)

        self.game_center_widget = QWidget.createWindowContainer(
            self.game_center_view, parent=self
        )
        self.resize(960, 540)
        self.game_center_widget.resize(960, 540)
        # self.game_center_widget.setFocus()
        # self.game_center_widget.setFocusPolicy(Qt.TabFocus)

        self.game_center_view.requestActivate()
        self.setFocus()

    def on_quit(self):
        self.close()



import os

import fsui
from fsbc.paths import Paths
from fsgs.FSGSDirectories import FSGSDirectories
from .launcher_config import LauncherConfig
from .launcher_settings import LauncherSettings
from .ui.Constants import Constants


class GamePaths(object):
    @staticmethod
    def current():
        model = LauncherConfig.get("amiga_model")
        if model.startswith("CD32"):
            platform = "CD32"
        elif model == "CDTV":
            platform = "CDTV"
        else:
            platform = "Amiga"
        name = LauncherSettings.get("config_name")
        uuid = LauncherConfig.get("x_game_uuid")
        return GamePaths(name, platform, uuid)

    def __init__(self, name, platform, uuid):
        self.uuid = uuid

        self.config_name = name
        if "(" in name:
            parts = name.split("(", 1)
            self.name, self.variant = parts
            self.name = self.name.strip()
            self.variant = self.variant.strip()
            if self.variant.endswith(")"):
                self.variant = self.variant[:-1]
            self.variant = self.variant.replace(") (", ", ")
            self.variant = self.variant.replace(")(", ", ")
        else:
            self.name = name
            self.variant = ""
        self.platform = platform

    def get_name(self):
        return self.name

    def get_variant(self):
        return self.variant

    @staticmethod
    def get_override_path(name):
        path = LauncherConfig.get(name)
        if not path:
            return ""
        path = Paths.expand_path(path)
        return path

    def get_screenshot_path(self, number):
        if number == 0:
            sha1 = LauncherConfig.get("title_sha1")
        else:
            sha1 = LauncherConfig.get("screen{0}_sha1".format(number))
        if sha1:
            return "sha1:" + sha1

        if number == 0:
            path = self.get_override_path("title_image")
        else:
            path = self.get_override_path("screen{0}_image".format(number))
        if path and os.path.exists(path):
            return path
        if self.uuid:
            if number == 0:
                name = "title.png"
            else:
                name = "screen{0}.png".format(number)
            paths = FSGSDirectories.get_images_dirs()
            for dir_ in paths:
                p = os.path.join(
                    dir_,
                    self.platform,
                    "Images",
                    self.uuid[:2],
                    self.uuid,
                    name,
                )
                if os.path.exists(p):
                    return p
                p = os.path.join(
                    dir_,
                    self.platform,
                    "Thumbnails",
                    self.uuid[:2],
                    self.uuid,
                    name,
                )
                if os.path.exists(p):
                    return p
        letter = self.get_letter(self.name)
        if not letter:
            return None
        name = self.name
        if number == 0:
            override_dir = LauncherConfig.get("titles_dir")
            if override_dir:
                paths = [Paths.expand_path(override_dir)]
            else:
                paths = FSGSDirectories.get_titles_dirs()
        else:
            override_dir = LauncherConfig.get("screenshots_dir")
            if override_dir:
                paths = [Paths.expand_path(override_dir)]
            else:
                paths = FSGSDirectories.get_screenshots_dirs()
        if number >= 2:
            name = "{0}_{1}".format(name, number)
        for dir_ in paths:
            path = os.path.join(dir_, letter, name + ".png")
            if os.path.exists(path):
                return path
            path = os.path.join(dir_, letter, name + ".gif")
            if os.path.exists(path):
                return path
            path = os.path.join(dir_, name + ".png")
            if os.path.exists(path):
                return path
            path = os.path.join(dir_, letter, name + ".gif")
            if os.path.exists(path):
                return path
        return None

    def load_screenshot(self, number):
        path = self.get_screenshot_path(number)
        if path:
            return fsui.Image(path)

    def load_screenshot_preview(self, number):
        image = self.load_screenshot(number)
        if image is None:
            return image
        if image.size == Constants.SCREEN_SIZE:
            return image
        if image.size[0] < 400:
            image.resize(
                (image.size[0] * 2, image.size[1] * 2), fsui.Image.NEAREST
            )
        image.resize(Constants.SCREEN_SIZE)
        return image

    def get_cover_path(self):
        sha1 = LauncherConfig.get("front_sha1")
        if sha1:
            return "sha1:" + sha1

        path = self.get_override_path("cover_image")
        if path and os.path.exists(path):
            return path
        if self.uuid:
            paths = FSGSDirectories.get_images_dirs()
            for dir_ in paths:
                p = os.path.join(
                    dir_,
                    self.platform,
                    "Images",
                    self.uuid[:2],
                    self.uuid,
                    "front.png",
                )
                if os.path.exists(p):
                    return p
                p = os.path.join(
                    dir_,
                    self.platform,
                    "Thumbnails",
                    self.uuid[:2],
                    self.uuid,
                    "front.png",
                )
                if os.path.exists(p):
                    return p
        letter = self.get_letter(self.name)
        if not letter:
            return None
        name = self.name
        override_dir = LauncherConfig.get("covers_dir")
        if override_dir:
            paths = [Paths.expand_path(override_dir)]
        else:
            paths = FSGSDirectories.get_covers_dirs()
        for dir_ in paths:
            path = os.path.join(dir_, letter, name + ".jpg")
            if os.path.exists(path):
                return path
            path = os.path.join(dir_, letter, name + ".png")
            if os.path.exists(path):
                return path
            path = os.path.join(dir_, name + ".jpg")
            if os.path.exists(path):
                return path
            path = os.path.join(dir_, name + ".png")
            if os.path.exists(path):
                return path
        return None

    def load_cover(self):
        path = self.get_cover_path()
        print(path)
        if path:
            return fsui.Image(path)

    def load_cover_preview(self):
        image = self.load_cover()
        if image is None:
            return image
        image.resize(Constants.COVER_SIZE)
        return image

    def get_theme_path(self):
        letter = self.get_letter(self.name)
        if not letter:
            return None
        paths = FSGSDirectories.get_themes_dirs()
        for dir_ in paths:
            path = os.path.join(dir_, letter, self.name)
            if os.path.exists(path):
                return path
        return None

    def _get_state_dir(self):
        config_name = self.config_name
        if not config_name:
            config_name = "Default"

        # use a temporary state dir, for now, to avoid problems with
        # floppy overlays etc interfering with net play
        from .netplay.netplay import Netplay

        if Netplay.current():
            # it is possible to manually specify the state dir
            config_name = LauncherConfig.get("__netplay_state_dir_name")
            if not config_name:
                # this is the default behavior, create a clean state
                # dir for the net play session
                netplay_game = LauncherConfig.get("__netplay_game")
                if netplay_game:
                    config_name = "Net Play ({0})".format(netplay_game)

        letter = self.get_letter(config_name)
        if not letter:
            config_name = "Default"
            letter = self.get_letter(config_name)
        # we use an existing state dir in a "letter" dir if it exists
        # (legacy support).
        path = os.path.join(
            FSGSDirectories.get_save_states_dir(), letter, config_name
        )
        if os.path.exists(path):
            return path
        # if not, we use a direct sub-folder of save states dir
        path = os.path.join(FSGSDirectories.get_save_states_dir(), config_name)
        return path

    def get_state_dir(self):
        state_dir = self._get_state_dir()
        if not os.path.exists(state_dir):
            os.makedirs(state_dir)
        return state_dir

    @staticmethod
    def get_letter(name):
        letter_name = name.upper()
        if letter_name.startswith("THE "):
            letter_name = letter_name[4:]
        if letter_name.startswith("A "):
            letter_name = letter_name[2:]
        for i in range(len(letter_name)):
            letter = letter_name[i]
            if letter in "01234567890":
                letter = "0"
                break
            if letter in "ABCDEFGHIJKLMNOPQRSTUVWXYZ":
                break
        else:
            return None
        return letter



import fsui
from launcher.i18n import gettext
from .netplay_panel import NetplayPanel


class NetplayDialog(fsui.Window):
    """Net play window. Max one window of this class should be open."""

    @classmethod
    def open(cls, parent=None):
        return fsui.open_window_instance(cls, parent)

    def __init__(self, parent=None):
        super().__init__(
            parent, "{} - FS-UAE Launcher".format(gettext("Net Play"))
        )
        self.layout = fsui.VerticalLayout()
        self.panel = NetplayPanel(self, header=False)
        self.layout.add(self.panel, expand=True, fill=True)

        self.panel.set_min_size((800, 500))
        self.panel.on_show()



from fsbc.system import macosx
from fsgs.drivers.gamedriver import Port
from fsgs.input.device import Device
from fsgs.input.keyboard import Keyboard


class InputMapper(object):
    def __init__(self, port, mapping, multiple=True):
        self.port = port  # type: Port
        self.device = port.device  # type: Device
        self.mapping = mapping
        self.multiple = multiple
        print(
            "[INPUT] Port",
            port.number,
            "Device:",
            self.device.id if self.device is not None else "(none)",
        )
        print("[INPUT] Port", port.number, "mapping:", mapping)

    def items(self):
        # input = self.input
        # if not input.device_config:
        #     return
        if not self.device:
            return
        config = self.device.configure(
            self.port.mapping_name, multiple=self.multiple
        )
        # print(config)
        print("mapping:", self.mapping)
        for native_button, game_button in config.items():
            game_button = game_button.upper()
            print("******", native_button, game_button)
            try:
                input_key = self.mapping[game_button]
            except KeyError:
                pass
            else:
                print("input_key", input_key)
                input_value = self.calc_input(native_button)
                # print("---->", input_key, input_value)
                # f.write('{input_key} = "{input_value}"\n'.format(
                #         input_key=input_key,
                #         input_value=input_value))
                if input_value is not None:
                    yield input_key, input_value

    def axis(self, axis, positive):
        raise NotImplementedError()

    def hat(self, hat, direction):
        raise NotImplementedError()

    def button(self, button):
        raise NotImplementedError()

    def key(self, key):
        raise NotImplementedError()

    def mouse(self, button, axis, positive):
        # raise NotImplementedError()
        return None

    def calc_input(self, value):
        parts = value.lower().split("_")
        if parts[0] == "axis":
            axis = int(parts[1])
            positive = parts[2] == "pos"
            return self.axis(axis, positive)
        elif parts[0] == "hat":
            hat = int(parts[1])
            direction = parts[2]
            return self.hat(hat, direction)
        elif parts[0] == "button":
            button = int(parts[1])
            return self.button(button)
        elif parts[0] == "key":
            key_name = value.split("_", 1)[1].lower()
            if key_name == "rctrl" and macosx:
                print("using ralt instead of rctrl on Mac")
                key_name = "ralt"
            key = Keyboard.key(key_name)
            return self.key(key)
        elif parts[0] == "mouse":
            button = 0
            axis = 0
            positive = False
            if parts[1] == "left":
                button = 1
            elif parts[1] == "right":
                button = 2
            elif parts[1] == "middle":
                button = 3
            else:
                axis = int(parts[1])
                positive = parts[2] == "pos"
            return self.mouse(button, axis, positive)
        return None
        # raise Exception("Invalid input value: {0}".format(value))


# class Key(object):
#     def __init__(self, name):
#         self.name = name

#     @property
#     def sdl_name(self):
#         return self.name

#     @property
#     def sdl_code(self):
#         return key_table[self.name][0]

#     @property
#     def dinput_code(self):
#         return key_table[self.name][2]

#     @property
#     def dinput_name(self):
#         return key_table[self.name][1]


# class Keyboard(object):
#     @staticmethod
#     def key(name):
#         if isinstance(name, int):
#             name = sdl_key_code_to_name[name]
#         name = name.upper()
#         if name.startswith("SDLK_"):
#             pass
#         else:
#             name = "SDLK_" + name
#         return Key(name)


# key_table = {
#     "SDLK_0": (48, "DIK_0", 11),
#     "SDLK_1": (49, "DIK_1", 2),
#     "SDLK_2": (50, "DIK_2", 3),
#     "SDLK_3": (51, "DIK_3", 4),
#     "SDLK_4": (52, "DIK_4", 5),
#     "SDLK_5": (53, "DIK_5", 6),
#     "SDLK_6": (54, "DIK_6", 7),
#     "SDLK_7": (55, "DIK_7", 8),
#     "SDLK_8": (56, "DIK_8", 9),
#     "SDLK_9": (57, "DIK_9", 10),
#     "SDLK_A": (97, "DIK_A", 30),
#     "SDLK_AMPERSAND": (38, "", 0),
#     "SDLK_ASTERISK": (42, "", 0),
#     "SDLK_AT": (64, "DIK_AT", 145),
#     "SDLK_B": (98, "DIK_B", 48),
#     "SDLK_BACKQUOTE": (96, "DIK_GRAVE", 0),
#     "SDLK_BACKSLASH": (92, "DIK_BACKSLASH", 43),
#     "SDLK_BACKSPACE": (8, "DIK_BACK", 14),
#     "SDLK_BREAK": (318, "", 0),
#     "SDLK_C": (99, "DIK_C", 46),
#     "SDLK_CAPSLOCK": (301, "DIK_CAPITAL", 58),
#     "SDLK_CARET": (94, "", 0),
#     "SDLK_CLEAR": (12, "", 0),
#     "SDLK_COLON": (58, "DIK_COLON", 146),
#     "SDLK_COMMA": (44, "DIK_COMMA", 51),
#     "SDLK_COMPOSE": (314, "", 0),
#     "SDLK_D": (100, "DIK_D", 32),
#     "SDLK_DELETE": (127, "DIK_DELETE", 211),
#     "SDLK_DOLLAR": (36, "", 0),
#     "SDLK_DOWN": (274, "DIK_DOWN", 208),
#     "SDLK_E": (101, "DIK_E", 18),
#     "SDLK_END": (279, "DIK_END", 207),
#     "SDLK_EQUALS": (61, "DIK_EQUALS", 13),
#     "SDLK_ESCAPE": (27, "DIK_ESCAPE", 1),
#     "SDLK_EURO": (321, "", 0),
#     "SDLK_EXCLAIM": (33, "", 0),
#     "SDLK_F": (102, "DIK_F", 33),
#     "SDLK_F1": (282, "DIK_F1", 59),
#     "SDLK_F10": (291, "DIK_F10", 68),
#     "SDLK_F11": (292, "DIK_F11", 87),
#     "SDLK_F12": (293, "DIK_F12", 88),
#     "SDLK_F13": (294, "DIK_F13", 100),
#     "SDLK_F14": (295, "DIK_F14", 101),
#     "SDLK_F15": (296, "DIK_F15", 102),
#     "SDLK_F2": (283, "DIK_F2", 60),
#     "SDLK_F3": (284, "DIK_F3", 61),
#     "SDLK_F4": (285, "DIK_F4", 62),
#     "SDLK_F5": (286, "DIK_F5", 63),
#     "SDLK_F6": (287, "DIK_F6", 64),
#     "SDLK_F7": (288, "DIK_F7", 65),
#     "SDLK_F8": (289, "DIK_F8", 66),
#     "SDLK_F9": (290, "DIK_F9", 67),
#     "SDLK_G": (103, "DIK_G", 34),
#     "SDLK_GREATER": (62, "", 0),
#     "SDLK_H": (104, "DIK_H", 35),
#     "SDLK_HASH": (35, "", 0),
#     "SDLK_HELP": (315, "", 0),
#     "SDLK_HOME": (278, "DIK_HOME", 199),
#     "SDLK_I": (105, "DIK_I", 23),
#     "SDLK_INSERT": (277, "DIK_INSERT", 210),
#     "SDLK_J": (106, "DIK_J", 36),
#     "SDLK_K": (107, "DIK_K", 37),
#     "SDLK_KP0": (256, "DIK_NUMPAD0", 82),
#     "SDLK_KP1": (257, "DIK_NUMPAD1", 79),
#     "SDLK_KP2": (258, "DIK_NUMPAD2", 80),
#     "SDLK_KP3": (259, "DIK_NUMPAD3", 81),
#     "SDLK_KP4": (260, "DIK_NUMPAD4", 75),
#     "SDLK_KP5": (261, "DIK_NUMPAD5", 76),
#     "SDLK_KP6": (262, "DIK_NUMPAD6", 77),
#     "SDLK_KP7": (263, "DIK_NUMPAD7", 71),
#     "SDLK_KP8": (264, "DIK_NUMPAD8", 72),
#     "SDLK_KP9": (265, "DIK_NUMPAD9", 73),
#     "SDLK_KP_DIVIDE": (267, "DIK_DIVIDE", 181),
#     "SDLK_KP_ENTER": (271, "DIK_NUMPADENTER", 156),
#     "SDLK_KP_EQUALS": (272, "", 141),
#     "SDLK_KP_MINUS": (269, "DIK_SUBTRACT", 74),
#     "SDLK_KP_MULTIPLY": (268, "DIK_MULTIPLY", 55),
#     "SDLK_KP_PERIOD": (266, "DIK_DECIMAL", 83),
#     "SDLK_KP_PLUS": (270, "DIK_ADD", 78),
#     "SDLK_L": (108, "DIK_L", 38),
#     "SDLK_LALT": (308, "DIK_LMENU", 56),
#     "SDLK_LCTRL": (306, "DIK_LCONTROL", 29),
#     "SDLK_LEFT": (276, "DIK_LEFT", 203),
#     "SDLK_LEFTBRACKET": (91, "DIK_LBRACKET", 26),
#     "SDLK_LEFTPAREN": (40, "", 0),
#     "SDLK_LESS": (60, "", 0),
#     "SDLK_LMETA": (310, "", 0),
#     "SDLK_LSHIFT": (304, "DIK_LSHIFT", 42),
#     "SDLK_LSUPER": (311, "DIK_LWIN", 219),
#     "SDLK_M": (109, "DIK_M", 50),
#     "SDLK_MENU": (319, "", 0),
#     "SDLK_MINUS": (45, "DIK_MINUS", 12),
#     "SDLK_MODE": (313, "", 0),
#     "SDLK_N": (110, "DIK_N", 49),
#     "SDLK_NUMLOCK": (300, "DIK_NUMLOCK", 69),
#     "SDLK_O": (111, "DIK_O", 24),
#     "SDLK_P": (112, "DIK_P", 25),
#     "SDLK_PAGEDOWN": (281, "DIK_NEXT", 209),
#     "SDLK_PAGEUP": (280, "DIK_PRIOR", 201),
#     "SDLK_PAUSE": (19, "", 0),
#     "SDLK_PERIOD": (46, "DIK_PERIOD", 52),
#     "SDLK_PLUS": (43, "", 0),
#     "SDLK_POWER": (320, "", 0),
#     "SDLK_PRINT": (316, "", 0),
#     "SDLK_Q": (113, "DIK_Q", 16),
#     "SDLK_QUESTION": (63, "", 0),
#     "SDLK_QUOTE": (39, "DIK_APOSTROPHE", 40),
#     "SDLK_QUOTEDBL": (34, "", 0),
#     "SDLK_R": (114, "DIK_R", 19),
#     "SDLK_RALT": (307, "DIK_RMENU", 184),
#     "SDLK_RCTRL": (305, "DIK_RCONTROL", 157),
#     "SDLK_RETURN": (13, "DIK_RETURN", 28),
#     "SDLK_RIGHT": (275, "DIK_RIGHT", 205),
#     "SDLK_RIGHTBRACKET": (93, "DIK_RBRACKET", 27),
#     "SDLK_RIGHTPAREN": (41, "", 0),
#     "SDLK_RMETA": (309, "", 0),
#     "SDLK_RSHIFT": (303, "DIK_RSHIFT", 54),
#     "SDLK_RSUPER": (312, "DIK_RWIN", 220),
#     "SDLK_S": (115, "DIK_S", 31),
#     "SDLK_SCROLLOCK": (302, "DIK_SCROLL", 70),
#     "SDLK_SEMICOLON": (59, "DIK_SEMICOLON", 39),
#     "SDLK_SLASH": (47, "DIK_SLASH", 53),
#     "SDLK_SPACE": (32, "DIK_SPACE", 57),
#     "SDLK_SYSREQ": (317, "DIK_SYSRQ", 183),
#     "SDLK_T": (116, "DIK_T", 20),
#     "SDLK_TAB": (9, "DIK_TAB", 15),
#     "SDLK_U": (117, "DIK_U", 22),
#     "SDLK_UNDERSCORE": (95, "", 147),
#     "SDLK_UNDO": (322, "", 0),
#     "SDLK_UP": (273, "DIK_UP", 200),
#     "SDLK_V": (118, "DIK_V", 47),
#     "SDLK_W": (119, "DIK_W", 17),
#     "SDLK_WORLD_0": (160, ""),
#     "SDLK_WORLD_1": (161, ""),
#     "SDLK_WORLD_10": (170, ""),
#     "SDLK_WORLD_11": (171, ""),
#     "SDLK_WORLD_12": (172, ""),
#     "SDLK_WORLD_13": (173, ""),
#     "SDLK_WORLD_14": (174, ""),
#     "SDLK_WORLD_15": (175, ""),
#     "SDLK_WORLD_16": (176, ""),
#     "SDLK_WORLD_17": (177, ""),
#     "SDLK_WORLD_18": (178, ""),
#     "SDLK_WORLD_19": (179, ""),
#     "SDLK_WORLD_2": (162, ""),
#     "SDLK_WORLD_20": (180, ""),
#     "SDLK_WORLD_21": (181, ""),
#     "SDLK_WORLD_22": (182, ""),
#     "SDLK_WORLD_23": (183, ""),
#     "SDLK_WORLD_24": (184, ""),
#     "SDLK_WORLD_25": (185, ""),
#     "SDLK_WORLD_26": (186, ""),
#     "SDLK_WORLD_27": (187, ""),
#     "SDLK_WORLD_28": (188, ""),
#     "SDLK_WORLD_29": (189, ""),
#     "SDLK_WORLD_3": (163, ""),
#     "SDLK_WORLD_30": (190, ""),
#     "SDLK_WORLD_31": (191, ""),
#     "SDLK_WORLD_32": (192, ""),
#     "SDLK_WORLD_33": (193, ""),
#     "SDLK_WORLD_34": (194, ""),
#     "SDLK_WORLD_35": (195, ""),
#     "SDLK_WORLD_36": (196, ""),
#     "SDLK_WORLD_37": (197, ""),
#     "SDLK_WORLD_38": (198, ""),
#     "SDLK_WORLD_39": (199, ""),
#     "SDLK_WORLD_4": (164, ""),
#     "SDLK_WORLD_40": (200, ""),
#     "SDLK_WORLD_41": (201, ""),
#     "SDLK_WORLD_42": (202, ""),
#     "SDLK_WORLD_43": (203, ""),
#     "SDLK_WORLD_44": (204, ""),
#     "SDLK_WORLD_45": (205, ""),
#     "SDLK_WORLD_46": (206, ""),
#     "SDLK_WORLD_47": (207, ""),
#     "SDLK_WORLD_48": (208, ""),
#     "SDLK_WORLD_49": (209, ""),
#     "SDLK_WORLD_5": (165, ""),
#     "SDLK_WORLD_50": (210, ""),
#     "SDLK_WORLD_51": (211, ""),
#     "SDLK_WORLD_52": (212, ""),
#     "SDLK_WORLD_53": (213, ""),
#     "SDLK_WORLD_54": (214, ""),
#     "SDLK_WORLD_55": (215, ""),
#     "SDLK_WORLD_56": (216, ""),
#     "SDLK_WORLD_57": (217, ""),
#     "SDLK_WORLD_58": (218, ""),
#     "SDLK_WORLD_59": (219, ""),
#     "SDLK_WORLD_6": (166, ""),
#     "SDLK_WORLD_60": (220, ""),
#     "SDLK_WORLD_61": (221, ""),
#     "SDLK_WORLD_62": (222, ""),
#     "SDLK_WORLD_63": (223, ""),
#     "SDLK_WORLD_64": (224, ""),
#     "SDLK_WORLD_65": (225, ""),
#     "SDLK_WORLD_66": (226, ""),
#     "SDLK_WORLD_67": (227, ""),
#     "SDLK_WORLD_68": (228, ""),
#     "SDLK_WORLD_69": (229, ""),
#     "SDLK_WORLD_7": (167, ""),
#     "SDLK_WORLD_70": (230, ""),
#     "SDLK_WORLD_71": (231, ""),
#     "SDLK_WORLD_72": (232, ""),
#     "SDLK_WORLD_73": (233, ""),
#     "SDLK_WORLD_74": (234, ""),
#     "SDLK_WORLD_75": (235, ""),
#     "SDLK_WORLD_76": (236, ""),
#     "SDLK_WORLD_77": (237, ""),
#     "SDLK_WORLD_78": (238, ""),
#     "SDLK_WORLD_79": (239, ""),
#     "SDLK_WORLD_8": (168, ""),
#     "SDLK_WORLD_80": (240, ""),
#     "SDLK_WORLD_81": (241, ""),
#     "SDLK_WORLD_82": (242, ""),
#     "SDLK_WORLD_83": (243, ""),
#     "SDLK_WORLD_84": (244, ""),
#     "SDLK_WORLD_85": (245, ""),
#     "SDLK_WORLD_86": (246, ""),
#     "SDLK_WORLD_87": (247, ""),
#     "SDLK_WORLD_88": (248, ""),
#     "SDLK_WORLD_89": (249, ""),
#     "SDLK_WORLD_9": (169, ""),
#     "SDLK_WORLD_90": (250, ""),
#     "SDLK_WORLD_91": (251, ""),
#     "SDLK_WORLD_92": (252, ""),
#     "SDLK_WORLD_93": (253, ""),
#     "SDLK_WORLD_94": (254, ""),
#     "SDLK_WORLD_95": (255, ""),
#     "SDLK_X": (120, "DIK_X", 45),
#     "SDLK_Y": (121, "DIK_Y", 21),
#     "SDLK_Z": (122, "DIK_Z", 44),
# }

# sdl_key_code_to_name = {}
# for __key, __value in key_table.items():
#     sdl_key_code_to_name[__value[0]] = __key
